# Copyright (C) 2003-2022, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

# Execute tests on a real FTP server (other tests use mock code).
#
# These tests write some files and directories on the local client and the
# remote server. You'll need write access in the login directory. These tests
# can take a few minutes because they have to wait to test the timezone
# calculation.

import datetime
import ftplib
import functools
import gc
import operator
import os
import pathlib
import time
import stat

import pytest

import ftputil.error
import ftputil.file_transfer
import ftputil.path_encoding
import ftputil.session
import ftputil.stat_cache

import test


def expected_time_shift():
    """
    Return the expected time shift in seconds.

    Unfortunately, the calculation may depend on the timezone of the server,
    i. e. the timezone used in directory listings coming from the server.

    So, depending on your test environment, you may need to change this
    function for your environment. _If_ you need an adapted
    `expected_time_shift`, please contact me (my e-mail address is in the
    ftputil documentation).

    In my particular case, I use Pure-FTPd as FTP server for the integration
    tests. At some point, it returned listings in the local timezone of the
    server, later it used UTC time, and now it uses the local timezone again.
    I wasn't able to find out why or how I can control this.
    """
    raw_time_shift = (datetime.datetime.now() - datetime.datetime.utcnow()).seconds
    # To be safe, round the above value to units of 900 s (1/4 hours).
    return round(raw_time_shift / 900.0) * 900


# The containerized PureFTPd seems to use UTC.
EXPECTED_TIME_SHIFT = 0.0


# Set when starting container
PORT = 2121

DEFAULT_SESSION_FACTORY = ftputil.session.session_factory(
    port=PORT, encrypt_data_channel=False
)


class Cleaner:
    """
    This class helps remove directories and files which might otherwise be left
    behind if a test fails in unexpected ways.
    """

    def __init__(self, host):
        # The test class (probably `RealFTPTest`) and the helper
        # class share the same `FTPHost` object.
        self._host = host
        self._ftp_items = []

    def add_dir(self, path):
        """
        Schedule a directory with path `path` for removal.
        """
        self._ftp_items.append(("d", self._host.path.abspath(path)))

    def add_file(self, path):
        """
        Schedule a file with path `path` for removal.
        """
        self._ftp_items.append(("f", self._host.path.abspath(path)))

    def clean(self):
        """
        Remove the directories and files previously remembered. The removal
        works in reverse order of the scheduling with `add_dir` and `add_file`.

        Errors due to a removal are ignored.
        """
        self._host.chdir("/")
        for type_, path in reversed(self._ftp_items):
            try:
                if type_ == "d":
                    # If something goes wrong in `rmtree` we might leave a mess
                    # behind.
                    self._host.rmtree(path)
                elif type_ == "f":
                    # Minor mess if `remove` fails
                    self._host.remove(path)
            except ftputil.error.FTPError:
                pass


class RealFTPTest:
    def setup_method(self, method):
        # Server, username, password.
        self.login_data = ("localhost", "ftptest", "dummy")
        self.host = ftputil.FTPHost(
            *self.login_data, session_factory=DEFAULT_SESSION_FACTORY
        )
        self.cleaner = Cleaner(self.host)

    def teardown_method(self, method):
        self.cleaner.clean()
        self.host.close()

    #
    # Helper methods
    #
    def make_remote_file(self, path):
        """Create a file on the FTP host."""
        self.cleaner.add_file(path)
        with self.host.open(path, "wb") as file_:
            # Write something. Otherwise the FTP server might not update the
            # time of last modification if the file existed before.
            file_.write(b"\n")

    def make_local_file(self):
        """
        Create a file on the local host (= on the client side).
        """
        with open("_local_file_", "wb") as fobj:
            fobj.write(b"abc\x12\x34def\t")


class TestMkdir(RealFTPTest):
    def test_directory_exists_after_mkdir(self):
        """
        After a `mkdir` call, the directory should exist.
        """
        host = self.host
        dir_name = "_testdir_"
        file_name = host.path.join(dir_name, "_nonempty_")
        self.cleaner.add_dir(dir_name)
        # Make dir and check if the directory is there.
        host.mkdir(dir_name)
        files = host.listdir(host.curdir)
        assert dir_name in files

    def test_makedirs_without_existing_dirs(self):
        """
        If some directories in a `makedirs` call don't exist yet, they should
        be implcitly created.
        """
        host = self.host
        # No `_dir1_` yet
        assert "_dir1_" not in host.listdir(host.curdir)
        # Vanilla case, all should go well.
        host.makedirs("_dir1_/dir2/dir3/dir4")
        self.cleaner.add_dir("_dir1_")
        # Check host.
        assert host.path.isdir("_dir1_")
        assert host.path.isdir("_dir1_/dir2")
        assert host.path.isdir("_dir1_/dir2/dir3")
        assert host.path.isdir("_dir1_/dir2/dir3/dir4")

    def test_makedirs_from_non_root_directory(self):
        # This is a testcase for issue #22, see
        # http://ftputil.sschwarzer.net/trac/ticket/22 .
        host = self.host
        # No `_dir1_` and `_dir2_` yet
        assert "_dir1_" not in host.listdir(host.curdir)
        assert "_dir2_" not in host.listdir(host.curdir)
        # Part 1: Try to make directories starting from `_dir1_` and
        # change to non-root directory.
        self.cleaner.add_dir("_dir1_")
        host.mkdir("_dir1_")
        host.chdir("_dir1_")
        host.makedirs("_dir2_/_dir3_")
        # Test for expected directory hierarchy.
        assert host.path.isdir("/_dir1_")
        assert host.path.isdir("/_dir1_/_dir2_")
        assert host.path.isdir("/_dir1_/_dir2_/_dir3_")
        assert not host.path.isdir("/_dir1_/_dir1_")
        # Remove all but the directory we're in.
        host.rmdir("/_dir1_/_dir2_/_dir3_")
        host.rmdir("/_dir1_/_dir2_")
        # Part 2: Try to make directories starting from root.
        self.cleaner.add_dir("/_dir2_")
        host.makedirs("/_dir2_/_dir3_")
        # Test for expected directory hierarchy
        assert host.path.isdir("/_dir2_")
        assert host.path.isdir("/_dir2_/_dir3_")
        assert not host.path.isdir("/_dir1_/_dir2_")

    def test_makedirs_of_existing_directory(self):
        """
        Calling `makedirs` on an existing directory with `exist_ok` should do
        nothing.
        """
        host = self.host
        # The (chrooted) login directory
        host.makedirs("/", exist_ok=True)

    def test_makedirs_with_file_in_the_way(self):
        """
        If a part of `makedirs`'s path is a file, a `PermanentError` should be
        raised.
        """
        host = self.host
        self.cleaner.add_dir("_dir1_")
        host.mkdir("_dir1_")
        self.make_remote_file("_dir1_/file1")
        # Try it.
        with pytest.raises(ftputil.error.PermanentError):
            host.makedirs("_dir1_/file1")
        with pytest.raises(ftputil.error.PermanentError):
            host.makedirs("_dir1_/file1/dir2")

    def test_makedirs_with_existing_directory(self):
        """
        An already-existent subdirectory in the path of `makedirs` should be
        ignored.
        """
        host = self.host
        self.cleaner.add_dir("_dir1_")
        host.mkdir("_dir1_")
        host.makedirs("_dir1_/dir2")
        # Check
        assert host.path.isdir("_dir1_")
        assert host.path.isdir("_dir1_/dir2")

    def test_makedirs_in_non_writable_directory(self):
        """
        If `makedirs` is asked to create a directory in a non-writable
        directory, a `PermanentError` should be raised.
        """
        host = self.host
        # Preparation: `rootdir1` exists but is only writable by root.
        with pytest.raises(ftputil.error.PermanentError):
            host.makedirs("rootdir1/dir2")

    def test_makedirs_with_writable_directory_at_end(self):
        """
        If `makedirs` is asked to create a directory in a writable directory
        under a non-writable directory, the `makedirs` call should succeed.
        """
        host = self.host
        self.cleaner.add_dir("rootdir2/dir2")
        # Preparation: `rootdir2` exists but is only writable by root. `dir2`
        # is writable by regular ftp users. Both directories below should work.
        host.makedirs("rootdir2/dir2", exist_ok=True)
        host.makedirs("rootdir2/dir2/dir3")
        assert host.path.isdir("rootdir2/dir2")
        assert host.path.isdir("rootdir2/dir2/dir3")


class TestRemoval(RealFTPTest):

    # Tests for `remove`

    def test_remove_existing_file(self):
        """
        Removal of an existing file should succeed.
        """
        self.cleaner.add_file("_testfile_")
        self.make_remote_file("_testfile_")
        host = self.host
        assert host.path.isfile("_testfile_")
        host.remove("_testfile_")
        assert not host.path.exists("_testfile_")

    def test_cache_invalidation_for_remove_exception(self):
        """
        If a file is removed, its stats information should be removed from the
        cache. This should also work if the removal raises an exception.
        """
        # Test for ticket #150
        self.cleaner.add_file("_testfile_")
        self.make_remote_file("_testfile_")
        host = self.host
        assert host.path.isfile("_testfile_")
        # Monkey-patch session `delete` call.
        old_delete = host._session.delete
        def failing_delete(path):
            # Simulate the case where the file is removed on the server, but a
            # proper reply doesn't get through to the client. It doesn't matter
            # whether the exception is `error_temp` or `error_perm`.
            old_delete(path)
            raise ftplib.error_perm("simulated error")

        host._session.delete = failing_delete
        with pytest.raises(ftputil.error.PermanentError):
            host.remove("_testfile_")
        assert not host.path.exists("_testfile_")

    def test_remove_non_existent_item(self):
        """
        If trying to remove a non-existent file system item, a `PermanentError`
        should be raised.
        """
        host = self.host
        with pytest.raises(ftputil.error.PermanentError):
            host.remove("nonexistent")

    def test_remove_on_directory(self):
        """
        Calling `remove` on a directory should raise a `PermanentError`.

        (Directories must be removed with `rmdir`.)
        """
        host = self.host
        dir_name = "_testdir_"
        self.cleaner.add_dir(dir_name)
        host.mkdir(dir_name)
        try:
            try:
                host.remove(dir_name)
            except ftputil.error.PermanentError as exc:
                assert str(exc).startswith("remove/unlink can only delete files")
            else:
                pytest.fail("we shouldn't have come here")
        finally:
            # Delete empty directory.
            host.rmdir(dir_name)
        files = host.listdir(host.curdir)
        assert dir_name not in files

    # Test for `rmdir`

    def test_remove_existing_directory(self):
        """
        If `rmdir` is called on an empty directory, the call should succeed.
        """
        self.cleaner.add_dir("_testdir_")
        host = self.host
        host.mkdir("_testdir_")
        assert host.path.isdir("_testdir_")
        host.rmdir("_testdir_")
        assert not host.path.exists("_testdir_")

    def test_rmdir_on_nonempty_directory(self):
        """
        If a directory exists, but isn't empty, `rmdir` should raise a
        `PermanentError`.
        """
        host = self.host
        dir_name = "_testdir_"
        self.cleaner.add_dir(dir_name)
        host.mkdir(dir_name)
        # Try to remove a non-empty directory.
        file_name = host.path.join(dir_name, "_nonempty_")
        self.cleaner.add_file(file_name)
        non_empty = host.open(file_name, "w")
        non_empty.close()
        with pytest.raises(ftputil.error.PermanentError):
            host.rmdir(dir_name)

    def test_cache_invalidation_for_rmdir_exception(self):
        """
        If a directory is removed, its stats information should be removed from
        the cache. This should also work if the removal raises an exception.
        """
        # Test for ticket #150
        self.cleaner.add_dir("_testdir_")
        host = self.host
        host.mkdir("_testdir_")
        assert host.path.isdir("_testdir_")
        # Monkey-patch session `rmd` call.
        old_rmd = host._session.rmd
        def failing_rmd(path):
            # Simulate the case where the directory is removed on the server,
            # but a proper reply doesn't get through to the client. It doesn't
            # matter whether the exception is `error_temp` or `error_perm`.
            old_rmd(path)
            raise ftplib.error_perm("simulated error")

        host._session.rmd = failing_rmd
        with pytest.raises(ftputil.error.PermanentError):
            host.rmdir("_testdir_")
        assert not host.path.exists("_testdir_")

    # Tests for `rmtree`

    def build_tree(self, host):
        """
        Build a directory tree for tests.
        """
        self.cleaner.add_dir("_dir1_")
        host.makedirs("_dir1_/dir2")
        self.make_remote_file("_dir1_/file1")
        self.make_remote_file("_dir1_/file2")
        self.make_remote_file("_dir1_/dir2/file3")
        self.make_remote_file("_dir1_/dir2/file4")

    def test_rmtree_without_error_handler(self):
        """
        Calling `rmtree` on a directory tree should remove the directory and
        everything under it.
        """
        host = self.host
        self.build_tree(host)
        # Remove `dir2`.
        host.rmtree("_dir1_/dir2")
        assert not host.path.exists("_dir1_/dir2")
        assert host.path.exists("_dir1_/file2")
        # Re-create `dir2` and remove `_dir1_`.
        host.mkdir("_dir1_/dir2")
        self.make_remote_file("_dir1_/dir2/file3")
        self.make_remote_file("_dir1_/dir2/file4")
        host.rmtree("_dir1_")
        assert not host.path.exists("_dir1_")

    def test_rmtree_with_error_handler(self):
        """
        If an `rmtree` call specifies an error handler, it should be used.
        """
        host = self.host
        self.cleaner.add_dir("_dir1_")
        host.mkdir("_dir1_")
        self.make_remote_file("_dir1_/file1")
        # Prepare error "handler"
        log = []

        def error_handler(*args):
            log.append(args)

        # Try to remove a file as root "directory".
        host.rmtree("_dir1_/file1", ignore_errors=True, onerror=error_handler)
        assert log == []
        host.rmtree("_dir1_/file1", ignore_errors=False, onerror=error_handler)
        assert log[0][0] == host.listdir
        assert log[0][1] == "_dir1_/file1"
        assert log[1][0] == host.rmdir
        assert log[1][1] == "_dir1_/file1"
        host.rmtree("_dir1_")
        # Try to remove a non-existent directory.
        del log[:]
        host.rmtree("_dir1_", ignore_errors=False, onerror=error_handler)
        assert log[0][0] == host.listdir
        assert log[0][1] == "_dir1_"
        assert log[1][0] == host.rmdir
        assert log[1][1] == "_dir1_"

    def test_remove_file_with_rmtree(self):
        """
        Calling `rmtree` on a file should raise a `PermanentError`.
        """
        host = self.host
        self.build_tree(host)
        with pytest.raises(ftputil.error.PermanentError):
            host.rmtree("_dir1_/file2")


class TestWalk(RealFTPTest):
    """
    Walk the directory tree

      walk_test
      ├── dir1
      │   ├── dir11
      │   └── dir12
      │       ├── dir123
      │       │   └── file1234
      │       ├── file121
      │       └── file122
      ├── dir2
      ├── dir3
      │   ├── dir31
      │   ├── dir32 -> ../dir1/dir12/dir123
      │   ├── file31
      │   └── file32
      └── file4

    and check if the results are the expected ones.
    """

    def _walk_test(self, expected_result, **walk_kwargs):
        """
        Walk the directory and test results.
        """
        # Collect data using `walk`.
        actual_result = []
        for items in self.host.walk(**walk_kwargs):
            actual_result.append(items)
        # Compare with expected results.
        assert len(actual_result) == len(expected_result)
        for index, _ in enumerate(actual_result):
            assert actual_result[index] == expected_result[index]

    def test_walk_topdown(self):
        # Preparation: build tree in directory `walk_test`.
        expected_result = [
            ("walk_test", ["dir1", "dir2", "dir3"], ["file4"]),
            #
            ("walk_test/dir1", ["dir11", "dir12"], []),
            #
            ("walk_test/dir1/dir11", [], []),
            #
            ("walk_test/dir1/dir12", ["dir123"], ["file121", "file122"]),
            #
            ("walk_test/dir1/dir12/dir123", [], ["file1234"]),
            #
            ("walk_test/dir2", [], []),
            #
            ("walk_test/dir3", ["dir31", "dir32"], ["file31", "file32"]),
            #
            ("walk_test/dir3/dir31", [], []),
        ]
        self._walk_test(expected_result, top="walk_test")

    def test_walk_depth_first(self):
        # Preparation: build tree in directory `walk_test`
        expected_result = [
            ("walk_test/dir1/dir11", [], []),
            #
            ("walk_test/dir1/dir12/dir123", [], ["file1234"]),
            #
            ("walk_test/dir1/dir12", ["dir123"], ["file121", "file122"]),
            #
            ("walk_test/dir1", ["dir11", "dir12"], []),
            #
            ("walk_test/dir2", [], []),
            #
            ("walk_test/dir3/dir31", [], []),
            #
            ("walk_test/dir3", ["dir31", "dir32"], ["file31", "file32"]),
            #
            ("walk_test", ["dir1", "dir2", "dir3"], ["file4"]),
        ]
        self._walk_test(expected_result, top="walk_test", topdown=False)

    def test_walk_following_links(self):
        # Preparation: build tree in directory `walk_test`.
        expected_result = [
            ("walk_test", ["dir1", "dir2", "dir3"], ["file4"]),
            #
            ("walk_test/dir1", ["dir11", "dir12"], []),
            #
            ("walk_test/dir1/dir11", [], []),
            #
            ("walk_test/dir1/dir12", ["dir123"], ["file121", "file122"]),
            #
            ("walk_test/dir1/dir12/dir123", [], ["file1234"]),
            #
            ("walk_test/dir2", [], []),
            #
            ("walk_test/dir3", ["dir31", "dir32"], ["file31", "file32"]),
            #
            ("walk_test/dir3/dir31", [], []),
            #
            ("walk_test/dir3/dir32", [], ["file1234"]),
        ]
        self._walk_test(expected_result, top="walk_test", followlinks=True)


class TestRename(RealFTPTest):
    def test_rename(self):
        """
        If `rename` is called on two paths, the first should disappear and the
        second should exist afterward.

        This test also tests `Path` support.
        """
        host = self.host
        # Make sure both files are gone after the test.
        self.cleaner.add_file("_testfile1_")
        self.cleaner.add_file("_testfile2_")
        self.make_remote_file("_testfile1_")
        host.rename(pathlib.Path("_testfile1_"), "_testfile2_")
        assert not host.path.exists("_testfile1_")
        assert host.path.exists(pathlib.Path("_testfile2_"))

    def test_cache_invalidation_for_rename(self):
        """
        A `rename` call should invalidate the cache entries for both path
        arguments. The purpose of the first invalidation is obvious, but the
        second invalidation is also important in case the path of the target
        existed before the rename and changes according to the new file stat
        information.

        This test also tests `Path` support.
        """
        # Test for ticket #150
        host = self.host
        # Make sure both files are gone after the test.
        self.cleaner.add_file("_testfile1_")
        self.cleaner.add_file("_testfile2_")
        # Case 1: Target file doesn't exist yet.
        self.make_remote_file("_testfile1_")
        file1_stat = host.stat("_testfile1_")
        host.rename(pathlib.Path("_testfile1_"), "_testfile2_")
        assert not host.path.exists("_testfile1_")
        assert host.path.exists(pathlib.Path("_testfile2_"))
        # Case 2: Target file already exists.
        #  Write the source file with a size different from the target file, so
        #  we can check whether we find the old or the new stat information
        #  when stat'ing the target file after the rename.
        with host.open("_testfile1_", "w") as fobj:
            fobj.write("abcdef\n")
        self.make_remote_file("_testfile2_")
        file1_stat = host.stat("_testfile1_")
        file2_stat = host.stat("_testfile2_")
        host.rename(pathlib.Path("_testfile1_"), "_testfile2_")
        assert not host.path.exists("_testfile1_")
        assert host.path.exists(pathlib.Path("_testfile2_"))
        new_file2_stat = host.stat("_testfile2_")
        assert new_file2_stat.st_size > file2_stat.st_size

    def test_cache_invalidation_for_rename_exception(self):
        """
        If a file system item is renamed/moved, its stats information should be
        removed from the cache. This should also work if the rename operation
        raises an exception.
        """
        # Test for ticket #150
        host = self.host
        # Make sure the target of the renaming operation is removed later.
        # Make sure both files are gone after the test.
        self.cleaner.add_file("_testfile1_")
        self.cleaner.add_file("_testfile2_")
        # Write the source file with a size different from the target file, so
        # we can check whether we find the old or the new stat information
        # when stat'ing the target file after the rename.
        with host.open("_testfile1_", "w") as fobj:
            fobj.write("abcdef\n")
        self.make_remote_file("_testfile2_")
        file1_stat = host.stat("_testfile1_")
        file2_stat = host.stat("_testfile2_")
        # Monkey-patch session `rename` call.
        old_rename = host._session.rename
        def failing_rename(source, target):
            # Simulate the case where the rename completely or partially
            # succeeds on the server, but a proper reply doesn't get through to
            # the client. It doesn't matter whether the exception is
            # `error_temp` or `error_perm`.
            old_rename(source, target)
            raise ftplib.error_perm("simulated error")
        host._session.rename = failing_rename
        #
        with pytest.raises(ftputil.error.PermanentError):
            host.rename(pathlib.Path("_testfile1_"), "_testfile2_")
        assert not host.path.exists("_testfile1_")
        assert host.path.exists(pathlib.Path("_testfile2_"))
        new_file2_stat = host.stat("_testfile2_")
        assert new_file2_stat.st_size > file2_stat.st_size

    def test_rename_with_spaces_in_directory(self):
        """
        `rename` should work if source and target contain a directory with
        spaces in its name.

        This test also tests `Path` support.
        """
        host = self.host
        dir_name = "_dir with spaces_"
        self.cleaner.add_dir(dir_name)
        host.mkdir(pathlib.Path(dir_name))
        self.make_remote_file(dir_name + "/testfile1")
        host.rename(dir_name + "/testfile1", pathlib.Path(dir_name + "/testfile2"))
        assert not host.path.exists(dir_name + "/testfile1")
        assert host.path.exists(dir_name + "/testfile2")


class TestStat(RealFTPTest):
    def test_stat(self):
        """
        Test some stat-related calls.
        """
        host = self.host
        dir_name = "_testdir_"
        file_name = host.path.join(dir_name, "_nonempty_")
        # Make a directory and a file in it.
        self.cleaner.add_dir(dir_name)
        host.mkdir(dir_name)
        with host.open(file_name, "wb") as fobj:
            fobj.write(b"abc\x12\x34def\t")
        # Do some stats
        # - dir
        dir_stat = host.stat(dir_name)
        assert isinstance(dir_stat._st_name, str)
        assert host.listdir(dir_name) == ["_nonempty_"]
        assert host.path.isdir(dir_name)
        assert not host.path.isfile(dir_name)
        assert not host.path.islink(dir_name)
        # - file
        file_stat = host.stat(file_name)
        assert isinstance(file_stat._st_name, str)
        assert not host.path.isdir(file_name)
        assert host.path.isfile(file_name)
        assert not host.path.islink(file_name)
        assert host.path.getsize(file_name) == 9
        # - file's modification time
        host.synchronize_times()
        #  The returned server mtime is supposed to be converted to UTC, so
        #  there should be only a small difference between server and client
        #  time. Arbitrarily allow two minutes here to account for limited time
        #  precision from parsing the directory.
        server_mtime = host.path.getmtime(file_name)
        client_mtime = (
            datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).timestamp()
        )
        assert not (client_mtime - server_mtime > 120)

    def test_issomething_for_nonexistent_directory(self):
        """
        If the `is*` methods are called on a path where the base directory
        doesn't exist, all of the methods should return `False`.
        """
        host = self.host
        # Check if we get the right results if even the containing directory
        # doesn't exist (see ticket #66).
        nonexistent_path = "/nonexistent/nonexistent"
        assert not host.path.isdir(nonexistent_path)
        assert not host.path.isfile(nonexistent_path)
        assert not host.path.islink(nonexistent_path)

    def test_special_broken_link(self):
        """
        If `is*` methods are called on a broken link (i.e. the link target
        doesn't exist), `islink` should return `True` and `isdir` and `isfile`
        should return `False`.
        """
        # Test for ticket #39.
        host = self.host
        broken_link_name = os.path.join("dir_with_broken_link", "nonexistent")
        assert host.lstat(broken_link_name)._st_target == "../nonexistent/nonexistent"
        assert not host.path.isdir(broken_link_name)
        assert not host.path.isfile(broken_link_name)
        assert host.path.islink(broken_link_name)

    def test_concurrent_access(self):
        """
        If we have two `FTPHost` instances for the same server directory,
        operations on one `FTPHost` instance shouldn't influence the cache
        entries of the other `FTPHost` instance.
        """
        self.make_remote_file("_testfile_")
        with ftputil.FTPHost(
            *self.login_data, session_factory=DEFAULT_SESSION_FACTORY
        ) as host1:
            with ftputil.FTPHost(
                *self.login_data, session_factory=DEFAULT_SESSION_FACTORY
            ) as host2:
                stat_result1 = host1.stat("_testfile_")
                stat_result2 = host2.stat("_testfile_")
                assert stat_result1 == stat_result2
                host2.remove("_testfile_")
                # Can still get the result via `host1`
                stat_result1 = host1.stat("_testfile_")
                assert stat_result1 == stat_result2
                # Stat'ing on `host2` gives an exception.
                with pytest.raises(ftputil.error.PermanentError):
                    host2.stat("_testfile_")
                # Stat'ing on `host1` after invalidation
                absolute_path = host1.path.join(host1.getcwd(), "_testfile_")
                host1.stat_cache.invalidate(absolute_path)
                with pytest.raises(ftputil.error.PermanentError):
                    host1.stat("_testfile_")

    def test_cache_auto_resizing(self):
        """
        Test if the cache is resized appropriately.
        """
        host = self.host
        cache = host.stat_cache._cache
        # Make sure the cache size isn't adjusted towards smaller values.
        unused_entries = host.listdir("walk_test")
        assert cache.size == ftputil.stat_cache.StatCache._DEFAULT_CACHE_SIZE
        # Make the cache very small initially and see if it gets resized.
        cache.size = 2
        entries = host.listdir("walk_test")
        # The adjusted cache size should be larger or equal to the number of
        # items in `walk_test` and its parent directory. The latter is read
        # implicitly upon `listdir`'s `isdir` call.
        expected_min_cache_size = max(len(host.listdir(host.curdir)), len(entries))
        assert cache.size >= expected_min_cache_size


class TestUploadAndDownload(RealFTPTest):
    """
    Test upload and download (including time shift test).
    """

    def test_time_shift(self):
        self.host.synchronize_times()
        assert self.host.time_shift() == EXPECTED_TIME_SHIFT

    @pytest.mark.slow_test
    def test_upload(self):
        """
        `upload_if_newer` should respect the time difference between client and
        server, including consideration of the time shift.
        """
        host = self.host
        host.synchronize_times()
        local_file = "_local_file_"
        remote_file = "_remote_file_"
        # Make local file to upload.
        self.make_local_file()
        # Wait, else small time differences between client and server actually
        # could trigger the update.
        time.sleep(65)
        try:
            self.cleaner.add_file(remote_file)
            host.upload(local_file, remote_file)
            # Retry; shouldn't be uploaded
            uploaded = host.upload_if_newer(local_file, remote_file)
            assert uploaded is False
            # Rewrite the local file.
            self.make_local_file()
            # Retry; should be uploaded now
            uploaded = host.upload_if_newer(local_file, remote_file)
            assert uploaded is True
        finally:
            # Clean up
            os.unlink(local_file)

    @pytest.mark.slow_test
    def test_download(self):
        """
        `download_if_newer` should respect the time difference between client
        and server, including consideration of the time shift.
        """
        host = self.host
        host.synchronize_times()
        local_file = "_local_file_"
        remote_file = "_remote_file_"
        # Make a remote file.
        self.make_remote_file(remote_file)
        # File should be downloaded as it's not present yet.
        downloaded = host.download_if_newer(remote_file, local_file)
        assert downloaded is True
        try:
            # If the remote file, taking the datetime precision into account,
            # _might_ be newer, the file will be downloaded again. To prevent
            # this, wait a bit over a minute (the remote precision), then
            # "touch" the local file.
            time.sleep(65)
            # Create empty file.
            with open(local_file, "w") as fobj:
                pass
            # Local file is present and newer, so shouldn't download.
            downloaded = host.download_if_newer(remote_file, local_file)
            assert downloaded is False
            # Re-make the remote file.
            self.make_remote_file(remote_file)
            # Local file is present but possibly older (taking the possible
            # deviation because of the precision into account), so should
            # download.
            downloaded = host.download_if_newer(remote_file, local_file)
            assert downloaded is True
        finally:
            # Clean up.
            os.unlink(local_file)

    def test_callback_with_transfer(self):
        """
        A `callback` argument should be called during a file transfer.
        """
        host = self.host
        FILE_NAME = "large_file"
        # Default chunk size as in `FTPHost.copyfileobj`
        MAX_COPY_CHUNK_SIZE = ftputil.file_transfer.MAX_COPY_CHUNK_SIZE
        file_size = host.path.getsize(FILE_NAME)
        chunk_count, _ = divmod(file_size, MAX_COPY_CHUNK_SIZE)
        # Add one chunk for remainder.
        chunk_count += 1
        # Define a callback that just collects all data passed to it.
        transferred_chunks_list = []

        def test_callback(chunk):
            transferred_chunks_list.append(chunk)

        try:
            host.download(FILE_NAME, FILE_NAME, callback=test_callback)
            # Construct a list of data chunks we expect.
            expected_chunks_list = []
            with open(FILE_NAME, "rb") as downloaded_fobj:
                while True:
                    chunk = downloaded_fobj.read(MAX_COPY_CHUNK_SIZE)
                    if not chunk:
                        break
                    expected_chunks_list.append(chunk)
            # Examine data collected by callback function.
            assert len(transferred_chunks_list) == chunk_count
            assert transferred_chunks_list == expected_chunks_list
        finally:
            os.unlink(FILE_NAME)


class TestFTPFiles(RealFTPTest):
    def test_only_closed_children(self):
        """
        If `FTPHost.open` needs an `FTPFile` object, an existing closed
        `FTPFile` should be reused.
        """
        REMOTE_FILE_NAME = "CONTENTS"
        host = self.host
        with host.open(REMOTE_FILE_NAME, "rb") as file_obj1:
            # Create empty file and close it.
            with host.open(REMOTE_FILE_NAME, "rb") as file_obj2:
                pass
            # This should re-use the second child because the first isn't
            # closed but the second is.
            with host.open(REMOTE_FILE_NAME, "rb") as file_obj:
                assert len(host._children) == 2
                assert file_obj._host is host._children[1]

    def test_no_timed_out_children(self):
        """
        `FTPHost.open` shouldn't use an `FTPFile` object that has timed out.
        """
        REMOTE_FILE_NAME = "CONTENTS"
        host = self.host
        # Implicitly create child host object.
        with host.open(REMOTE_FILE_NAME, "rb") as file_obj1:
            pass
        # Monkey-patch file to simulate an FTP server timeout below.
        def timed_out_pwd():
            raise ftplib.error_temp("simulated timeout")

        file_obj1._host._session.pwd = timed_out_pwd
        # Try to get a file - which shouldn't be the timed-out file.
        with host.open(REMOTE_FILE_NAME, "rb") as file_obj2:
            assert file_obj1 is not file_obj2
        # Re-use closed and not timed-out child session.
        with host.open(REMOTE_FILE_NAME, "rb") as file_obj3:
            pass
        assert file_obj2 is file_obj3

    def test_no_delayed_226_children(self):
        """
        `FTPHost.open` shouldn't use an `FTPFile` object that gave an 226 error
        reply.
        """
        REMOTE_FILE_NAME = "CONTENTS"
        host = self.host
        # Implicitly create child host object.
        with host.open(REMOTE_FILE_NAME, "rb") as file_obj1:
            pass
        # Monkey-patch file to simulate an FTP server timeout below.
        def timed_out_pwd():
            raise ftplib.error_reply("delayed 226 reply")

        file_obj1._host._session.pwd = timed_out_pwd
        # Try to get a file - which shouldn't be the timed-out file.
        with host.open(REMOTE_FILE_NAME, "rb") as file_obj2:
            assert file_obj1 is not file_obj2
        # Re-use closed and not timed-out child session.
        with host.open(REMOTE_FILE_NAME, "rb") as file_obj3:
            pass
        assert file_obj2 is file_obj3


class TestChmod(RealFTPTest):
    def assert_mode(self, path, expected_mode):
        """
        Return an integer containing the allowed bits in the mode change
        command.

        The `FTPHost` object to test against is `self.host`.
        """
        full_mode = self.host.stat(path).st_mode
        # Remove flags we can't set via `chmod`. Allowed flags according to
        # Python documentation: https://docs.python.org/library/stat.html
        allowed_flags = [
            stat.S_ISUID,
            stat.S_ISGID,
            stat.S_ENFMT,
            stat.S_ISVTX,
            stat.S_IREAD,
            stat.S_IWRITE,
            stat.S_IEXEC,
            stat.S_IRWXU,
            stat.S_IRUSR,
            stat.S_IWUSR,
            stat.S_IXUSR,
            stat.S_IRWXG,
            stat.S_IRGRP,
            stat.S_IWGRP,
            stat.S_IXGRP,
            stat.S_IRWXO,
            stat.S_IROTH,
            stat.S_IWOTH,
            stat.S_IXOTH,
        ]
        allowed_mask = functools.reduce(operator.or_, allowed_flags)
        mode = full_mode & allowed_mask
        assert mode == expected_mode, "mode {0:o} != {1:o}".format(mode, expected_mode)

    def test_chmod_existing_directory(self):
        """
        A `chmod` operation on a writable existing directory should succeed.
        """
        host = self.host
        host.mkdir("_test dir_")
        self.cleaner.add_dir("_test dir_")
        # Set/get mode of a directory.
        host.chmod("_test dir_", 0o757)
        self.assert_mode("_test dir_", 0o757)
        # Set/get mode in nested directory.
        host.mkdir("_test dir_/nested_dir")
        self.cleaner.add_dir("_test dir_/nested_dir")
        host.chmod("_test dir_/nested_dir", 0o757)
        self.assert_mode("_test dir_/nested_dir", 0o757)

    def test_chmod_existing_file(self):
        """
        A `chmod` operation on a writable existing file should succeed.
        """
        host = self.host
        host.mkdir("_test dir_")
        self.cleaner.add_dir("_test dir_")
        # Set/get mode on a file.
        file_name = host.path.join("_test dir_", "_testfile_")
        self.make_remote_file(file_name)
        host.chmod(file_name, 0o646)
        self.assert_mode(file_name, 0o646)

    def test_chmod_nonexistent_path(self):
        """
        If `chmod` is called on a non-existent path, a `PermanentError` should
        be raised.
        """
        # Set/get mode of a non-existing item.
        with pytest.raises(ftputil.error.PermanentError):
            self.host.chmod("nonexistent", 0o757)

    def test_cache_invalidation(self):
        """
        If the mode of a directory or file is changed, make sure that the cache
        doesn't contain stat entries before the `chmod` call.
        """
        host = self.host
        host.mkdir("_test dir_")
        self.cleaner.add_dir("_test dir_")
        # Make sure the mode is in the cache.
        unused_stat_result = host.stat("_test dir_")
        # Set/get mode of the directory.
        host.chmod("_test dir_", 0o757)
        self.assert_mode("_test dir_", 0o757)
        # Set/get mode on a file.
        file_name = host.path.join("_test dir_", "_testfile_")
        self.make_remote_file(file_name)
        # Make sure the mode is in the cache.
        unused_stat_result = host.stat(file_name)
        host.chmod(file_name, 0o646)
        self.assert_mode(file_name, 0o646)


class TestRestArgument(RealFTPTest):

    TEST_FILE_NAME = "rest_test"

    def setup_method(self, method):
        super().setup_method(method)
        # Write test file.
        with self.host.open(self.TEST_FILE_NAME, "wb") as fobj:
            fobj.write(b"abcdefghijkl")
        self.cleaner.add_file(self.TEST_FILE_NAME)

    def test_for_reading(self):
        """
        If a `rest` argument is passed to `open`, the following read operation
        should start at the byte given by `rest`.
        """
        with self.host.open(self.TEST_FILE_NAME, "rb", rest=3) as fobj:
            data = fobj.read()
        assert data == b"defghijkl"

    def test_for_writing(self):
        """
        If a `rest` argument is passed to `open`, the following write operation
        should start writing at the byte given by `rest`.
        """
        with self.host.open(self.TEST_FILE_NAME, "wb", rest=3) as fobj:
            fobj.write(b"123")
        with self.host.open(self.TEST_FILE_NAME, "rb") as fobj:
            data = fobj.read()
        assert data == b"abc123"

    def test_invalid_read_from_text_file(self):
        """
        If the `rest` argument is used for reading from a text file, a
        `CommandNotImplementedError` should be raised.
        """
        with pytest.raises(ftputil.error.CommandNotImplementedError):
            self.host.open(self.TEST_FILE_NAME, "r", rest=3)

    def test_invalid_write_to_text_file(self):
        """
        If the `rest` argument is used for reading from a text file, a
        `CommandNotImplementedError` should be raised.
        """
        with pytest.raises(ftputil.error.CommandNotImplementedError):
            self.host.open(self.TEST_FILE_NAME, "w", rest=3)

    # There are no tests for reading and writing beyond the end of a file. For
    # example, if the remote file is 10 bytes long and
    # `open(remote_file, "rb", rest=100)` is used, the server may return an
    # error status code or not.
    #
    # The server I use for testing returns a 554 status when attempting to
    # _read_ beyond the end of the file. On the other hand, if attempting to
    # _write_ beyond the end of the file, the server accepts the request, but
    # starts writing after the end of the file, i. e. appends to the file.
    #
    # Instead of expecting certain responses that may differ between server
    # implementations, I leave the bahavior for too large `rest` arguments
    # undefined. In practice, this shouldn't be a problem because the `rest`
    # argument should only be used for error recovery, and in this case a valid
    # byte count for the `rest` argument should be known.


class TestOther(RealFTPTest):
    def test_open_for_reading(self):
        # Test for issues #17 and #51,
        # http://ftputil.sschwarzer.net/trac/ticket/17 and
        # http://ftputil.sschwarzer.net/trac/ticket/51 .
        file1 = self.host.open("large_file", "rb")
        time.sleep(1)
        # Depending on the FTP server, this might return a status code
        # unexpected by `ftplib` or block the socket connection until a
        # server-side timeout.
        file1.close()

    def test_subsequent_reading(self):
        """
        An `FTPHost.open` call should reuse a closed `FTPFile`.
        """
        # Open a file for reading.
        with self.host.open("CONTENTS", "rb") as file1:
            pass
        # Make sure that there are no problems if the connection is reused.
        with self.host.open("CONTENTS", "rb") as file2:
            pass
        assert file1._session is file2._session

    def test_names_with_spaces(self):
        """
        Directory and file names with spaces in them shouldn't cause problems.

        (We've seen problems here with some servers before
        `_robust_ftp_command` was introduced.)
        """
        host = self.host
        assert host.path.isdir("dir with spaces")
        assert host.listdir("dir with spaces") == [
            "second dir",
            "some file",
            "some_file",
        ]
        assert host.path.isdir("dir with spaces/second dir")
        assert host.path.isfile("dir with spaces/some_file")
        assert host.path.isfile("dir with spaces/some file")

    def test_synchronize_times_without_write_access(self):
        """
        If the current directory isn't writable, `synchronize_times` should
        raise a `TimeShiftError`.
        """
        host = self.host
        # This isn't writable by the ftp account the tests are run under.
        host.chdir("rootdir1")
        with pytest.raises(ftputil.error.TimeShiftError):
            host.synchronize_times()

    def test_encoding(self):
        """
        If the session factory is requested with a specific encoding, the
        session from the factory should use this requested encoding.
        """
        for encoding in ["latin-1", "UTF-8"]:
            factory = ftputil.session.session_factory(port=PORT, encoding=encoding)
            session = factory(*self.login_data)
            assert session.encoding == encoding

    def test_listdir_with_non_ascii_byte_string(self):
        """
        `listdir` should accept byte strings with non-ASCII characters and
        return non-ASCII characters in directory or file names.
        """
        host = self.host
        path = "äbc".encode("UTF-8")
        names = host.listdir(path)
        assert names[0] == b"file1"
        assert names[1] == "file1_ö".encode("UTF-8")

    def test_listdir_with_non_ascii_unicode_string(self):
        """
        `listdir` should accept unicode strings with non-ASCII characters and
        return non-ASCII characters in directory or file names.
        """
        host = self.host
        # `ftplib` under Python 3 only works correctly if the unicode strings
        # are decoded from the `ftplib` default encoding. For Python 3.9 and up
        # the `encode`/`decode` combination is a no-op.
        path = "äbc".encode("UTF-8").decode(
            ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING
        )
        names = host.listdir(path)
        assert names[0] == "file1"
        assert names[1] == "file1_ö".encode("UTF-8").decode(
            ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING
        )

    @pytest.mark.skipif(
        ftputil.path_encoding.RUNNING_UNDER_PY39_AND_UP,
        reason="test applies only to `FTPHost` objects using 'latin-1' path encoding",
    )
    def test_path_with_non_latin1_unicode_string_below_python_3_9(self):
        """
        ftputil operations shouldn't accept file paths with non-latin1
        characters.
        """
        # Use some musical symbols. These are certainly not latin1. ;-)
        path = "𝄞𝄢"
        # `UnicodeEncodeError` is also the exception that `ftplib` raises if it
        # gets a non-latin1 path.
        with pytest.raises(UnicodeEncodeError):
            self.host.mkdir(path)

    def test_list_a_option(self):
        """
        If `use_list_a_option` is set to `False`, "hidden" files shouldn't be
        included in the `listdir` result.

        If `use_list_a_option` is set to `True`, "hidden" files should be
        included in the `listdir` result.

        Whether this works as described, depends on the FTP server and its
        configuration.
        """
        # For this test to pass, the server must _not_ list "hidden" files by
        # default but instead only when the `LIST` `-a` option is used.
        host = self.host
        assert not host.use_list_a_option
        directory_entries = host.listdir(host.curdir)
        assert ".hidden" not in directory_entries
        # Switch on showing of hidden paths.
        host.use_list_a_option = True
        directory_entries = host.listdir(host.curdir)
        assert ".hidden" in directory_entries

    def _make_objects_to_be_garbage_collected(self):
        for _ in range(10):
            with ftputil.FTPHost(
                *self.login_data, session_factory=DEFAULT_SESSION_FACTORY
            ) as host:
                for _ in range(10):
                    unused_stat_result = host.stat("CONTENTS")
                    with host.open("CONTENTS") as fobj:
                        unused_data = fobj.read()

    def test_garbage_collection(self):
        """
        Test whether there are cycles which prevent garbage collection.
        """
        gc.collect()
        objects_before_test = len(gc.garbage)
        self._make_objects_to_be_garbage_collected()
        gc.collect()
        objects_after_test = len(gc.garbage)
        assert not objects_after_test - objects_before_test



# Copyright (C) 2021, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
Compatibility code for Python versions <= 3.8 vs. >= 3.9

Python 3.9 changed the default encoding for FTP remote paths from latin-1 to
UTF-8 which causes a few headaches for ftputil.
"""

import sys


__all__ = ["DEFAULT_ENCODING", "RUNNING_UNDER_PY39_AND_UP", "FTPLIB_DEFAULT_ENCODING"]


RUNNING_UNDER_PY39_AND_UP = (sys.version_info.major, sys.version_info.minor) >= (3, 9)

# FTP path default encoding for Python 3.8 and below
FTPLIB_PY38_ENCODING = "latin-1"

# FTP path default encoding for Python 3.9 and above
FTPLIB_PY39_ENCODING = "utf-8"

# Default encoding for ftputil. Stay compatible to the behavior of former
# ftputil versions and Python <= 3.8.
DEFAULT_ENCODING = FTPLIB_PY38_ENCODING

if RUNNING_UNDER_PY39_AND_UP:
    FTPLIB_DEFAULT_ENCODING = FTPLIB_PY39_ENCODING
else:
    FTPLIB_DEFAULT_ENCODING = FTPLIB_PY38_ENCODING



# Copyright (C) 2002-2022, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
`FTPHost` is the central class of the `ftputil` library.

See `__init__.py` for an example.
"""

import datetime
import errno
import ftplib
import stat
import sys
import time

import ftputil.error
import ftputil.file
import ftputil.file_transfer
import ftputil.path
import ftputil.path_encoding
import ftputil.session
import ftputil.stat
import ftputil.tool


__all__ = ["FTPHost"]


# The "protected" attributes PyLint talks about aren't intended for clients of
# the library. `FTPHost` objects need to use some of these library-internal
# attributes though.
# pylint: disable=protected-access


# For Python versions 3.8 and below, ftputil has implicitly defaulted to
# latin-1 encoding. Prefer that behavior for Python 3.9 and up as well instead
# of using the encoding that is the default for `ftplib.FTP` in the Python
# version.
if ftputil.path_encoding.RUNNING_UNDER_PY39_AND_UP:

    class default_session_factory(ftplib.FTP):
        def __init__(self, *args, **kwargs):
            # Python 3.9 defines `encoding` as a keyword-only argument, so test
            # only for the `encoding` argument in `kwargs`.
            #
            # Only use the ftputil default encoding if the caller didn't pass
            # an encoding.
            if "encoding" not in kwargs:
                kwargs["encoding"] = ftputil.path_encoding.DEFAULT_ENCODING
            super().__init__(*args, **kwargs)

else:
    default_session_factory = ftplib.FTP


#####################################################################
# `FTPHost` class with several methods similar to those of `os`


class FTPHost:
    """
    FTP host class.
    """

    # Implementation notes:
    #
    # Upon every request of a file (`FTPFile` object) a new FTP session is
    # created (or reused from a cache), leading to a child session of the
    # `FTPHost` object from which the file is requested.
    #
    # This is needed because opening an `FTPFile` will make the local session
    # object wait for the completion of the transfer. In fact, code like this
    # would block indefinitely, if the `RETR` request would be made on the
    # `_session` of the object host:
    #
    #   host = FTPHost(ftp_server, user, password)
    #   f = host.open("index.html")
    #   host.getcwd()   # would block!
    #
    # On the other hand, the initially constructed host object will store
    # references to already established `FTPFile` objects and reuse an
    # associated connection if its associated `FTPFile` has been closed.

    def __init__(self, *args, **kwargs):
        """
        Abstract initialization of `FTPHost` object.
        """
        # Store arguments for later operations.
        self._args = args
        self._kwargs = kwargs
        # XXX: Maybe put the following in a `reset` method.
        # The time shift setting shouldn't be reset though. Make a session
        # according to these arguments.
        self._session = self._make_session()
        # Simulate `os.path`.
        self.path = ftputil.path._Path(self)
        # lstat, stat, listdir services.
        self._stat = ftputil.stat._Stat(self)
        self.stat_cache = self._stat._lstat_cache
        self.stat_cache.enable()
        with ftputil.error.ftplib_error_to_ftp_os_error:
            current_dir = self._session.pwd()
            self._cached_current_dir = self.path.normpath(
                ftputil.tool.as_str_path(current_dir, encoding=self._encoding)
            )
        # Associated `FTPHost` objects for data transfer.
        self._children = []
        # This is only set to something else than `None` if this instance
        # represents an `FTPFile`.
        self._file = None
        # Now opened.
        self.closed = False
        # Set curdir, pardir etc. for the remote host. RFC 959 states that this
        # is, strictly speaking, dependent on the server OS but it seems to
        # work at least with Unix and Windows servers.
        self.curdir, self.pardir, self.sep = ".", "..", "/"
        # Set default time shift (used in `upload_if_newer` and
        # `download_if_newer`).
        self._time_shift = 0.0
        # Don't use `LIST -a` option by default. If the server doesn't
        # understand the `-a` option and interprets it as a path, the results
        # can be surprising. See ticket #110.
        self.use_list_a_option = False

    def keep_alive(self):
        """
        Try to keep the connection alive in order to avoid server timeouts.

        Note that this won't help if the connection has already timed out! In
        this case, `keep_alive` will raise an `TemporaryError`. (Actually, if
        you get a server timeout, the error - for a specific connection - will
        be permanent.)
        """
        # Warning: Don't call this method on `FTPHost` instances which
        # represent file transfers. This may fail in confusing ways.
        with ftputil.error.ftplib_error_to_ftp_os_error:
            # Ignore return value.
            self._session.pwd()

    #
    # Dealing with child sessions and file-like objects (rather low-level)
    #
    def _make_session(self):
        """
        Return a new session object according to the current state of this
        `FTPHost` instance.
        """
        # Don't modify original attributes below.
        args = self._args[:]
        kwargs = self._kwargs.copy()
        # If a session factory has been given on the instantiation of this
        # `FTPHost` object, use the same factory for this `FTPHost` object's
        # child sessions.
        factory = kwargs.pop("session_factory", default_session_factory)
        with ftputil.error.ftplib_error_to_ftp_os_error:
            session = factory(*args, **kwargs)
            if not hasattr(session, "encoding"):
                raise ftputil.error.NoEncodingError(
                    f"session instance {session!r} must have an `encoding` attribute"
                )
            self._encoding = session.encoding
        return session

    def _copy(self):
        """
        Return a copy of this `FTPHost` object.
        """
        # The copy includes a new session factory return value (aka session)
        # but doesn't copy the state of `self.getcwd()`.
        return self.__class__(*self._args, **self._kwargs)

    def _available_child(self):
        """
        Return an available (i. e. one whose `_file` object is closed and
        doesn't have a timed-out server connection) child (`FTPHost` object)
        from the pool of children or `None` if there aren't any.
        """
        # TODO: Currently timed-out child sessions aren't removed and may
        # collect over time. In very busy or long running processes, this might
        # slow down an application because the same stale child sessions have
        # to be processed again and again.
        for host in self._children:
            # Test for timeouts only after testing for a closed file:
            # - If a file isn't closed, save time; don't bother to access the
            #   remote server.
            # - If a file transfer on the child is in progress, requesting the
            #   directory is an invalid operation because of the way the FTP
            #   state machine works (see RFC 959).
            if host._file.closed:
                try:
                    host._session.pwd()
                # Under high load, a 226 status response from a previous
                # download may arrive too late, so that it's "seen" in the
                # `pwd` call. For now, skip the potential child session; it
                # will be considered again when `_available_child` is called
                # the next time.
                except ftplib.error_reply:
                    continue
                # Timed-out sessions raise `error_temp`.
                except ftplib.error_temp:
                    continue
                # The server may have closed the connection which may cause
                # `host._session.getline` to raise an `EOFError` (see ticket
                # #114).
                except EOFError:
                    continue
                # Under high load, there may be a socket read timeout during
                # the last FTP file `close` (see ticket #112). Note that a
                # socket timeout is quite different from an FTP session
                # timeout.
                except OSError:
                    continue
                else:
                    # Everything's ok; use this `FTPHost` instance.
                    return host
        # Be explicit.
        return None

    def open(
        self,
        path,
        mode="r",
        buffering=None,
        encoding=None,
        errors=None,
        newline=None,
        *,
        rest=None,
    ):
        """
        Return an open file(-like) object which is associated with this
        `FTPHost` object.

        The arguments `path`, `mode`, `buffering`, `encoding`, `errors` and
        `newlines` have the same meaning as for `open`. If `rest` is given as
        an integer,

        - reading will start at the byte (zero-based) `rest`
        - writing will overwrite the remote file from byte `rest`

        This method tries to reuse a child but will generate a new one if none
        is available.
        """
        # Support the same arguments as `open`.
        # pylint: disable=too-many-arguments
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        host = self._available_child()
        if host is None:
            host = self._copy()
            self._children.append(host)
            host._file = ftputil.file.FTPFile(host)
        basedir = self.getcwd()
        # Prepare for changing the directory (see whitespace workaround in
        # method `_dir`).
        if host.path.isabs(path):
            effective_path = path
        else:
            effective_path = host.path.join(basedir, path)
        effective_dir, effective_file = host.path.split(effective_path)
        try:
            # This will fail if the directory isn't accessible at all.
            host.chdir(effective_dir)
        except ftputil.error.PermanentError:
            # Similarly to a failed `file` in a local file system, raise an
            # `IOError`, not an `OSError`.
            raise ftputil.error.FTPIOError(
                "remote directory '{}' doesn't "
                "exist or has insufficient access rights".format(effective_dir)
            )
        host._file._open(
            effective_file,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            errors=errors,
            newline=newline,
            rest=rest,
        )
        if "w" in mode:
            # Invalidate cache entry because size and timestamps will change.
            self.stat_cache.invalidate(effective_path)
        return host._file

    def close(self):
        """
        Close host connection.
        """
        if self.closed:
            return
        # Close associated children.
        for host in self._children:
            # Children have a `_file` attribute which is an `FTPFile` object.
            host._file.close()
            host.close()
        # Now deal with ourself.
        try:
            with ftputil.error.ftplib_error_to_ftp_os_error:
                self._session.close()
        finally:
            # If something went wrong before, the host/session is probably
            # defunct and subsequent calls to `close` won't help either, so
            # consider the host/session closed for practical purposes.
            self.stat_cache.clear()
            self._children = []
            self.closed = True

    #
    # Setting a custom directory parser
    #
    def set_parser(self, parser):
        """
        Set the parser for extracting stat results from directory listings.

        The parser interface is described in the documentation, but here are
        the most important things:

        - A parser should derive from `ftputil.stat.Parser`.

        - The parser has to implement two methods, `parse_line` and
          `ignores_line`. For the latter, there's a probably useful default
          in the class `ftputil.stat.Parser`.

        - `parse_line` should try to parse a line of a directory listing and
          return a `ftputil.stat.StatResult` instance. If parsing isn't
          possible, raise `ftputil.error.ParserError` with a useful error
          message.

        - `ignores_line` should return a true value if the line isn't assumed
          to contain stat information.
        """
        # The cache contents, if any, probably aren't useful.
        self.stat_cache.clear()
        # Set the parser explicitly, don't allow "smart" switching anymore.
        self._stat._parser = parser
        self._stat._allow_parser_switching = False

    #
    # Time shift adjustment between client (i. e. us) and server
    #
    @staticmethod
    def __rounded_time_shift(time_shift):
        """
        Return the given time shift in seconds, but rounded to 15-minute units.
        The argument is also assumed to be given in seconds.
        """
        minute = 60.0
        # Avoid division by zero below.
        if time_shift == 0:
            return 0.0
        # Use a positive value for rounding.
        absolute_time_shift = abs(time_shift)
        signum = time_shift / absolute_time_shift
        # Round absolute time shift to 15-minute units.
        absolute_rounded_time_shift = int(
            (absolute_time_shift + (7.5 * minute)) / (15.0 * minute)
        ) * (15.0 * minute)
        # Return with correct sign.
        return signum * absolute_rounded_time_shift

    def __assert_valid_time_shift(self, time_shift):
        """
        Perform sanity checks on the time shift value (given in seconds). If
        the value is invalid, raise a `TimeShiftError`, else simply return
        `None`.
        """
        minute = 60.0  # seconds
        hour = 60.0 * minute
        absolute_rounded_time_shift = abs(self.__rounded_time_shift(time_shift))
        # Test 1: Fail if the absolute time shift is greater than a full day
        #         (24 hours).
        if absolute_rounded_time_shift > 24 * hour:
            raise ftputil.error.TimeShiftError(
                "time shift abs({0:.2f} s) > 1 day".format(time_shift)
            )
        # Test 2: Fail if the deviation between given time shift and 15-minute
        #         units is greater than a certain limit.
        maximum_deviation = 5 * minute
        if abs(time_shift - self.__rounded_time_shift(time_shift)) > maximum_deviation:
            raise ftputil.error.TimeShiftError(
                "time shift ({0:.2f} s) deviates more than {1:d} s "
                "from 15-minute units".format(time_shift, int(maximum_deviation))
            )

    def set_time_shift(self, time_shift):
        """
        Set the time shift value.

        By (my) definition, the time shift value is the difference of
        the time zone used in server listings and UTC, i. e.

            time_shift =def= t_server - utc
        <=> t_server = utc + time_shift
        <=> utc = t_server - time_shift

        The time shift is measured in seconds.
        """
        self.__assert_valid_time_shift(time_shift)
        old_time_shift = self.time_shift()
        if time_shift != old_time_shift:
            # If the time shift changed, all entries in the cache will have
            # wrong times with respect to the updated time shift, therefore
            # clear the cache.
            self.stat_cache.clear()
            self._time_shift = self.__rounded_time_shift(time_shift)

    def time_shift(self):
        """
        Return the time shift between FTP server and client. See the
        docstring of `set_time_shift` for more on this value.
        """
        return self._time_shift

    def synchronize_times(self):
        """
        Synchronize the local times of FTP client and server. This is necessary
        to let `upload_if_newer` and `download_if_newer` work correctly. If
        `synchronize_times` isn't applicable (see below), the time shift can
        still be set explicitly with `set_time_shift`.

        This implementation of `synchronize_times` requires _all_ of the
        following:

        - The connection between server and client is established.
        - The client has write access to the directory that is current when
          `synchronize_times` is called.

        The common usage pattern of `synchronize_times` is to call it directly
        after the connection is established. (As can be concluded from the
        points above, this requires write access to the login directory.)

        If `synchronize_times` fails, it raises a `TimeShiftError`.
        """
        helper_file_name = "_ftputil_sync_"
        # Open a dummy file for writing in the current directory on the FTP
        # host, then close it.
        try:
            # May raise `FTPIOError` if directory isn't writable.
            file_ = self.open(helper_file_name, "w")
            file_.close()
        except ftputil.error.FTPIOError:
            raise ftputil.error.TimeShiftError(
                """couldn't write helper file in directory '{}'""".format(self.getcwd())
            )
        # If everything worked up to here it should be possible to stat and
        # then remove the just-written file.
        try:
            server_time = self.path.getmtime(helper_file_name)
            self.unlink(helper_file_name)
        except ftputil.error.FTPOSError:
            # If we got a `TimeShiftError` exception above, we should't come
            # here: if we didn't get a `TimeShiftError` above, deletion should
            # be possible. The only reason for an exception I can think of here
            # is a race condition by removing the helper file or write
            # permission from the directory or helper file after it has been
            # written to.
            raise ftputil.error.TimeShiftError(
                "could write helper file but not unlink it"
            )
        # Calculate the difference between server and client.
        now = time.time()
        time_shift = server_time - now
        # As the time shift for this host instance isn't set yet, the directory
        # parser will calculate times one year in the past if the time zone of
        # the server is east from ours. Thus the time shift will be off by a
        # year as well (see ticket #55).
        if time_shift < -360 * 24 * 60 * 60:
            # Read one year and recalculate the time shift. We don't know how
            # many days made up that year (it might have been a leap year), so
            # go the route via `datetime.replace`.
            server_datetime = datetime.datetime.fromtimestamp(
                server_time, tz=datetime.timezone.utc
            )
            server_datetime = server_datetime.replace(year=server_datetime.year + 1)
            time_shift = server_datetime.timestamp() - now
        self.set_time_shift(time_shift)

    #
    # Operations based on file-like objects (rather high-level), like upload
    # and download
    #
    # XXX: This has a different API from `shutil.copyfileobj`, on which this
    # method is modeled. But I don't think it makes sense to change this method
    # here because the method is probably rarely used and a change would break
    # client code.
    @staticmethod
    def copyfileobj(
        source,
        target,
        max_chunk_size=ftputil.file_transfer.MAX_COPY_CHUNK_SIZE,
        callback=None,
    ):
        """
        Copy data from file-like object `source` to file-like object `target`.
        """
        ftputil.file_transfer.copyfileobj(source, target, max_chunk_size, callback)

    def _upload_files(self, source_path, target_path):
        """
        Return a `LocalFile` and `RemoteFile` as source and target,
        respectively.

        The strings `source_path` and `target_path` are the (absolute or
        relative) paths of the local and the remote file, respectively.
        """
        source_file = ftputil.file_transfer.LocalFile(source_path, "rb")
        # Passing `self` (the `FTPHost` instance) here is correct.
        target_file = ftputil.file_transfer.RemoteFile(self, target_path, "wb")
        return source_file, target_file

    def upload(self, source, target, callback=None):
        """
        Upload a file from the local source (name) to the remote target (name).

        If a callable `callback` is given, it's called after every chunk of
        transferred data. The chunk size is a constant defined in
        `file_transfer`. The callback will be called with a single argument,
        the data chunk that was transferred before the callback was called.
        """
        if source in ["", b""]:
            raise IOError("path argument `source` is empty")
        ftputil.tool.raise_for_empty_path(target, path_argument_name="target")
        target = ftputil.tool.as_str_path(target, encoding=self._encoding)
        source_file, target_file = self._upload_files(source, target)
        ftputil.file_transfer.copy_file(
            source_file, target_file, conditional=False, callback=callback
        )

    def upload_if_newer(self, source, target, callback=None):
        """
        Upload a file only if it's newer than the target on the remote host or
        if the target file does not exist. See the method `upload` for the
        meaning of the parameters.

        If an upload was necessary, return `True`, else return `False`.

        If a callable `callback` is given, it's called after every chunk of
        transferred data. The chunk size is a constant defined in
        `file_transfer`. The callback will be called with a single argument,
        the data chunk that was transferred before the callback was called.
        """
        ftputil.tool.raise_for_empty_path(source, path_argument_name="source")
        if target in ["", b""]:
            raise IOError("path argument `target` is empty")
        target = ftputil.tool.as_str_path(target, encoding=self._encoding)
        source_file, target_file = self._upload_files(source, target)
        return ftputil.file_transfer.copy_file(
            source_file, target_file, conditional=True, callback=callback
        )

    def _download_files(self, source_path, target_path):
        """
        Return a `RemoteFile` and `LocalFile` as source and target,
        respectively.

        The strings `source_path` and `target_path` are the (absolute or
        relative) paths of the remote and the local file, respectively.
        """
        source_file = ftputil.file_transfer.RemoteFile(self, source_path, "rb")
        target_file = ftputil.file_transfer.LocalFile(target_path, "wb")
        return source_file, target_file

    def download(self, source, target, callback=None):
        """
        Download a file from the remote source (name) to the local target
        (name).

        If a callable `callback` is given, it's called after every chunk of
        transferred data. The chunk size is a constant defined in
        `file_transfer`. The callback will be called with a single argument,
        the data chunk that was transferred before the callback was called.
        """
        ftputil.tool.raise_for_empty_path(source, path_argument_name="source")
        if target in ["", b""]:
            raise IOError("path argument `target` is empty")
        source = ftputil.tool.as_str_path(source, encoding=self._encoding)
        source_file, target_file = self._download_files(source, target)
        ftputil.file_transfer.copy_file(
            source_file, target_file, conditional=False, callback=callback
        )

    def download_if_newer(self, source, target, callback=None):
        """
        Download a file only if it's newer than the target on the local host or
        if the target file does not exist. See the method `download` for the
        meaning of the parameters.

        If a download was necessary, return `True`, else return `False`.

        If a callable `callback` is given, it's called after every chunk of
        transferred data. The chunk size is a constant defined in
        `file_transfer`. The callback will be called with a single argument,
        the data chunk that was transferred before the callback was called.
        """
        if source in ["", b""]:
            raise IOError("path argument `source` is empty")
        ftputil.tool.raise_for_empty_path(target, path_argument_name="target")
        source = ftputil.tool.as_str_path(source, encoding=self._encoding)
        source_file, target_file = self._download_files(source, target)
        return ftputil.file_transfer.copy_file(
            source_file, target_file, conditional=True, callback=callback
        )

    #
    # Helper methods to descend into a directory before executing a command
    #
    def _check_inaccessible_login_directory(self):
        """
        Raise an `InaccessibleLoginDirError` exception if we can't change to
        the login directory. This test is only reliable if the current
        directory is the login directory.
        """
        presumable_login_dir = self.getcwd()
        # Bail out with an internal error rather than modify the current
        # directory without hope of restoration.
        try:
            self.chdir(presumable_login_dir)
        except ftputil.error.PermanentError:
            raise ftputil.error.InaccessibleLoginDirError(
                "directory '{}' is not accessible".format(presumable_login_dir)
            )

    def _robust_ftp_command(self, command, path, descend_deeply=False):
        """
        Run an FTP command on a path. The return value of the method is the
        return value of the command.

        If `descend_deeply` is true (the default is false), descend deeply,
        i. e. change the directory to the end of the path.
        """
        # If we can't change to the yet-current directory, the code below won't
        # work (see below), so in this case rather raise an exception than
        # giving wrong results.
        self._check_inaccessible_login_directory()
        # Some FTP servers don't behave as expected if the directory portion of
        # the path contains whitespace; some even yield strange results if the
        # command isn't executed in the current directory. Therefore, change to
        # the directory which contains the item to run the command on and
        # invoke the command just there.
        #
        # Remember old working directory.
        old_dir = self.getcwd()
        try:
            if descend_deeply:
                # Invoke the command in (not: on) the deepest directory.
                self.chdir(path)
                # Workaround for some servers that give recursive listings when
                # called with a dot as path; see issue #33,
                # http://ftputil.sschwarzer.net/trac/ticket/33
                return command(self, "")
            else:
                # Invoke the command in the "next to last" directory.
                head, tail = self.path.split(path)
                self.chdir(head)
                return command(self, tail)
        finally:
            self.chdir(old_dir)

    #
    # Miscellaneous utility methods resembling functions in `os`
    #
    def getcwd(self):
        """
        Return the current directory path.
        """
        return self._cached_current_dir

    def chdir(self, path):
        """
        Change the directory on the host to `path`.
        """
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        with ftputil.error.ftplib_error_to_ftp_os_error:
            self._session.cwd(path)
        # The path given as the argument is relative to the old current
        # directory, therefore join them.
        self._cached_current_dir = self.path.normpath(
            self.path.join(self._cached_current_dir, path)
        )

    # Ignore unused argument `mode`
    # pylint: disable=unused-argument
    def mkdir(self, path, mode=None):
        """
        Make the directory path on the remote host. The argument `mode` is
        ignored and only "supported" for similarity with `os.mkdir`.
        """
        ftputil.tool.raise_for_empty_path(path)
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)

        def command(self, path):
            """Callback function."""
            with ftputil.error.ftplib_error_to_ftp_os_error:
                self._session.mkd(path)

        self._robust_ftp_command(command, path)

    # TODO: The virtual directory support doesn't have unit tests yet because
    # the mocking most likely would be quite complicated. The tests should be
    # added when mainly the `mock` library is used instead of the mock code in
    # `test.mock_ftplib`.
    #
    # Ignore unused argument `mode`
    # pylint: disable=unused-argument
    def makedirs(self, path, mode=None, exist_ok=False):
        """
        Make the directory `path`, but also make not yet existing intermediate
        directories, like `os.makedirs`.

        The value of `mode` is only accepted for compatibility with
        `os.makedirs` but otherwise ignored.

        If `exist_ok` is `False` (the default) and the leaf directory exists,
        raise a `PermanentError` with `errno` 17.
        """
        ftputil.tool.raise_for_empty_path(path)
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        path = self.path.abspath(path)
        directories = path.split(self.sep)
        old_dir = self.getcwd()
        try:
            # Try to build the directory chain from the "uppermost" to the
            # "lowermost" directory.
            for index in range(1, len(directories)):
                # Re-insert the separator which got lost by using `path.split`.
                next_directory = self.sep + self.path.join(*directories[: index + 1])
                # If we have "virtual directories" (see #86), just listing the
                # parent directory won't tell us if a directory actually
                # exists. So try to change into the directory.
                try:
                    self.chdir(next_directory)
                except ftputil.error.PermanentError:
                    # Directory presumably doesn't exist.
                    try:
                        self.mkdir(next_directory)
                    except ftputil.error.PermanentError:
                        # Find out the cause of the error. Re-raise the
                        # exception only if the directory didn't exist already,
                        # else something went _really_ wrong, e. g. there's a
                        # regular file with the name of the directory.
                        if not self.path.isdir(next_directory):
                            raise
                else:
                    # Directory exists. If we are at the last directory
                    # component and `exist_ok` is `False`, this is an error.
                    if (index == len(directories) - 1) and (not exist_ok):
                        # Before PEP 3151, if `exist_ok` is `False`, trying to
                        # create an existing directory in the local file system
                        # results in an `OSError` with `errno.EEXIST, so
                        # emulate this also for FTP.
                        ftp_os_error = ftputil.error.PermanentError(
                            "path {!r} exists".format(path)
                        )
                        ftp_os_error.errno = errno.EEXIST
                        raise ftp_os_error
        finally:
            self.chdir(old_dir)

    def rmdir(self, path):
        """
        Remove the _empty_ directory `path` on the remote host.

        Compatibility note:

        Previous versions of ftputil could possibly delete non-empty
        directories as well, - if the server allowed it. This is no longer
        supported.
        """
        ftputil.tool.raise_for_empty_path(path)
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        path = self.path.abspath(path)
        if self.listdir(path):
            raise ftputil.error.PermanentError("directory '{}' not empty".format(path))
        # XXX: How does `rmd` work with links?
        def command(self, path):
            """Callback function."""
            with ftputil.error.ftplib_error_to_ftp_os_error:
                self._session.rmd(path)

        # Always invalidate the cache. If `_robust_ftp_command` raises an
        # exception, we can't tell for sure if the removal failed on the server
        # vs. it succeeded, but something went wrong after that.
        try:
            self._robust_ftp_command(command, path)
        finally:
            self.stat_cache.invalidate(path)

    def remove(self, path):
        """
        Remove the file or link given by `path`.

        Raise a `PermanentError` if the path doesn't exist, but maybe raise
        other exceptions depending on the state of the server (e. g. timeout).
        """
        ftputil.tool.raise_for_empty_path(path)
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        path = self.path.abspath(path)
        # Though `isfile` includes also links to files, `islink` is needed to
        # include links to directories.
        if (
            self.path.isfile(path)
            or self.path.islink(path)
            or not self.path.exists(path)
        ):
            # If the path doesn't exist, let the removal command trigger an
            # exception with a more appropriate error message.
            def command(self, path):
                """Callback function."""
                with ftputil.error.ftplib_error_to_ftp_os_error:
                    self._session.delete(path)

            # Always invalidate the cache. If `_robust_ftp_command` raises an
            # exception, we can't tell for sure if the removal failed on the
            # server vs. it succeeded, but something went wrong after that.
            try:
                self._robust_ftp_command(command, path)
            finally:
                self.stat_cache.invalidate(path)
        else:
            raise ftputil.error.PermanentError(
                "remove/unlink can only delete files and links, " "not directories"
            )

    unlink = remove

    def rmtree(self, path, ignore_errors=False, onerror=None):
        """
        Remove the given remote, possibly non-empty, directory tree. The
        interface of this method is rather complex, in favor of compatibility
        with `shutil.rmtree`.

        If `ignore_errors` is set to a true value, errors are ignored. If
        `ignore_errors` is a false value _and_ `onerror` isn't set, all
        exceptions occurring during the tree iteration and processing are
        raised. These exceptions are all of type `PermanentError`.

        To distinguish between error situations, pass in a callable for
        `onerror`. This callable must accept three arguments: `func`, `path`
        and `exc_info`. `func` is a bound method object, _for example_
        `your_host_object.listdir`. `path` is the path that was the recent
        argument of the respective method (`listdir`, `remove`, `rmdir`).
        `exc_info` is the exception info as it's got from `sys.exc_info`.

        Implementation note: The code is copied from `shutil.rmtree` in
        Python 2.4 and adapted to ftputil.
        """
        ftputil.tool.raise_for_empty_path(path)
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        # The following code is an adapted version of Python 2.4's
        # `shutil.rmtree` function.
        if ignore_errors:

            def new_onerror(*args):
                """Do nothing."""
                # pylint: disable=unused-argument
                pass

        elif onerror is None:

            def new_onerror(*args):
                """Re-raise exception."""
                # pylint: disable=misplaced-bare-raise, unused-argument
                raise

        else:
            new_onerror = onerror
        names = []
        try:
            names = self.listdir(path)
        except ftputil.error.PermanentError:
            new_onerror(self.listdir, path, sys.exc_info())
        for name in names:
            full_name = self.path.join(path, name)
            try:
                mode = self.lstat(full_name).st_mode
            except ftputil.error.PermanentError:
                mode = 0
            if stat.S_ISDIR(mode):
                self.rmtree(full_name, ignore_errors, new_onerror)
            else:
                try:
                    self.remove(full_name)
                except ftputil.error.PermanentError:
                    new_onerror(self.remove, full_name, sys.exc_info())
        try:
            self.rmdir(path)
        except ftputil.error.FTPOSError:
            new_onerror(self.rmdir, path, sys.exc_info())

    def rename(self, source, target):
        """
        Rename the `source` on the FTP host to `target`.
        """
        ftputil.tool.raise_for_empty_path(source, path_argument_name="source")
        ftputil.tool.raise_for_empty_path(target, path_argument_name="target")
        source = ftputil.tool.as_str_path(source, encoding=self._encoding)
        target = ftputil.tool.as_str_path(target, encoding=self._encoding)
        source_head, source_tail = self.path.split(source)
        target_head, target_tail = self.path.split(target)
        # Avoid code duplication below.
        #
        # Use `source_arg` and `target_arg` instead of `source` and `target` to
        # make it clearer that we use the arguments passed to
        # `rename_with_cleanup`, not any variables from the scope outside
        # `rename_with_cleanup`.
        def rename_with_cleanup(source_arg, target_arg):
            try:
                with ftputil.error.ftplib_error_to_ftp_os_error:
                    self._session.rename(source_arg, target_arg)
            # Always invalidate the cache entries in case the rename succeeds
            # on the server, but the server doesn't manage to tell the client.
            finally:
                source_absolute_path = self.path.abspath(source_arg)
                target_absolute_path = self.path.abspath(target_arg)
                self.stat_cache.invalidate(source_absolute_path)
                self.stat_cache.invalidate(target_absolute_path)

        # The following code is in spirit similar to the code in the method
        # `_robust_ftp_command`, though we do _not_ do _everything_ imaginable.
        self._check_inaccessible_login_directory()
        paths_contain_whitespace = (" " in source_head) or (" " in target_head)
        if paths_contain_whitespace and source_head == target_head:
            # Both items are in the same directory.
            old_dir = self.getcwd()
            try:
                self.chdir(source_head)
                rename_with_cleanup(source_tail, target_tail)
            finally:
                self.chdir(old_dir)
        else:
            # Use straightforward command.
            rename_with_cleanup(source, target)

    # XXX: One could argue to put this method into the `_Stat` class, but I
    # refrained from that because then `_Stat` would have to know about
    # `FTPHost`'s `_session` attribute and in turn about `_session`'s `dir`
    # method.
    def _dir(self, path):
        """
        Return a directory listing as made by FTP's `LIST` command as a list of
        strings.
        """
        # Don't use `self.path.isdir` in this method because that would cause a
        # call of `(l)stat` and thus a call to `_dir`, so we would end up with
        # an infinite recursion.
        def _FTPHost_dir_command(self, path):
            """Callback function."""
            lines = []

            def callback(line):
                """Callback function."""
                lines.append(ftputil.tool.as_str(line, encoding=self._encoding))

            with ftputil.error.ftplib_error_to_ftp_os_error:
                if self.use_list_a_option:
                    self._session.dir("-a", path, callback)
                else:
                    self._session.dir(path, callback)
            return lines

        lines = self._robust_ftp_command(
            _FTPHost_dir_command, path, descend_deeply=True
        )
        return lines

    # The `listdir`, `lstat` and `stat` methods don't use `_robust_ftp_command`
    # because they implicitly already use `_dir` which actually uses
    # `_robust_ftp_command`.

    def listdir(self, path):
        """
        Return a list of directories, files etc. in the directory named `path`.

        If the directory listing from the server can't be parsed with any of
        the available parsers raise a `ParserError`.
        """
        ftputil.tool.raise_for_empty_path(path)
        original_path = path
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        items = self._stat._listdir(path)
        return [
            ftputil.tool.same_string_type_as(original_path, item, self._encoding)
            for item in items
        ]

    def lstat(self, path, _exception_for_missing_path=True):
        """
        Return an object similar to that returned by `os.lstat`.

        If the directory listing from the server can't be parsed with any of
        the available parsers, raise a `ParserError`. If the directory _can_ be
        parsed and the `path` is _not_ found, raise a `PermanentError`.

        (`_exception_for_missing_path` is an implementation aid and _not_
        intended for use by ftputil clients.)
        """
        ftputil.tool.raise_for_empty_path(path)
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        return self._stat._lstat(path, _exception_for_missing_path)

    def stat(self, path, _exception_for_missing_path=True):
        """
        Return info from a "stat" call on `path`.

        If the directory containing `path` can't be parsed, raise a
        `ParserError`. If the directory containing `path` can be parsed but the
        `path` can't be found, raise a `PermanentError`. Also raise a
        `PermanentError` if there's an endless (cyclic) chain of symbolic links
        "behind" the `path`.

        (`_exception_for_missing_path` is an implementation aid and _not_
        intended for use by ftputil clients.)
        """
        ftputil.tool.raise_for_empty_path(path)
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        return self._stat._stat(path, _exception_for_missing_path)

    def walk(self, top, topdown=True, onerror=None, followlinks=False):
        """
        Iterate over directory tree and return a tuple (dirpath, dirnames,
        filenames) on each iteration, like the `os.walk` function (see
        https://docs.python.org/library/os.html#os.walk ).
        """
        ftputil.tool.raise_for_empty_path(top, path_argument_name="top")
        top = ftputil.tool.as_str_path(top, encoding=self._encoding)
        # The following code is copied from `os.walk` in Python 2.4 and adapted
        # to ftputil.
        try:
            names = self.listdir(top)
        except ftputil.error.FTPOSError as err:
            if onerror is not None:
                onerror(err)
            return
        dirs, nondirs = [], []
        for name in names:
            if self.path.isdir(self.path.join(top, name)):
                dirs.append(name)
            else:
                nondirs.append(name)
        if topdown:
            yield top, dirs, nondirs
        for name in dirs:
            path = self.path.join(top, name)
            if followlinks or not self.path.islink(path):
                yield from self.walk(path, topdown, onerror, followlinks)
        if not topdown:
            yield top, dirs, nondirs

    def chmod(self, path, mode):
        """
        Change the mode of a remote `path` (a string) to the integer `mode`.
        This integer uses the same bits as the mode value returned by the
        `stat` and `lstat` commands.

        If something goes wrong, raise a `TemporaryError` or a
        `PermanentError`, according to the status code returned by the server.
        In particular, a non-existent path usually causes a `PermanentError`.
        """
        ftputil.tool.raise_for_empty_path(path)
        path = ftputil.tool.as_str_path(path, encoding=self._encoding)
        path = self.path.abspath(path)

        def command(self, path):
            """Callback function."""
            with ftputil.error.ftplib_error_to_ftp_os_error:
                self._session.voidcmd("SITE CHMOD 0{0:o} {1}".format(mode, path))

        self._robust_ftp_command(command, path)
        self.stat_cache.invalidate(path)

    def __getstate__(self):
        raise TypeError("cannot serialize FTPHost object")

    #
    # Context manager methods
    #
    def __enter__(self):
        # Return `self`, so it can be accessed as the variable component of the
        # `with` statement.
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # We don't need the `exc_*` arguments here.
        # pylint: disable=unused-argument
        self.close()
        # Be explicit.
        return False



# Copyright (C) 2008-2019, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import ftplib
import io
import pytest

import ftputil.error

import test.scripted_session as scripted_session
from test import test_base


Call = scripted_session.Call


# Exception raised by client code, i. e. code using ftputil. Used to test the
# behavior in case of client exceptions.
class ClientCodeException(Exception):
    pass


#
# Test cases
#
class TestHostContextManager:
    def test_normal_operation(self):
        """
        If an `FTPHost` instance is created, it should be closed by the host
        context manager.
        """
        script = [
            Call(method_name="__init__"),
            Call(method_name="pwd", result="/"),
            Call(method_name="close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            assert host.closed is False
        assert host.closed is True

    def test_ftputil_exception(self):
        """
        If an `ftplib.FTP` method raises an exception, it should be caught by
        the host context manager and the host object should be closed.
        """
        script = [
            # Since `__init__` raises an exception, `pwd` isn't called.
            # However, `close` is called via the context manager.
            Call(method_name="__init__", result=ftplib.error_perm),
            Call(method_name="close"),
        ]
        with pytest.raises(ftputil.error.FTPOSError):
            with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
                pass
        # We arrived here, that's fine. Because the `FTPHost` object wasn't
        # successfully constructed, the assignment to `host` shouldn't have
        # happened.
        assert "host" not in locals()

    def test_client_code_exception(self):
        """
        If client code raises an exception in the context manager block, the
        host object should be closed by the context manager.
        """
        script = [
            Call(method_name="__init__"),
            Call(method_name="pwd", result="/"),
            Call(method_name="close"),
        ]
        try:
            with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
                assert host.closed is False
                raise ClientCodeException()
        except ClientCodeException:
            assert host.closed is True
        else:
            pytest.fail("`ClientCodeException` not raised")


class TestFileContextManager:
    def test_normal_operation(self):
        """
        If an `FTPFile` object is created in the `FTPFile` context manager, the
        context manager should close the file at the end of the `with` block.
        """
        host_script = [
            Call(method_name="__init__"),
            Call(method_name="pwd", result="/"),
            Call(method_name="close"),
        ]
        file_script = [
            Call(method_name="__init__"),
            Call(method_name="pwd", result="/"),
            Call(method_name="cwd", result=None, args=("/",)),
            Call(method_name="voidcmd", result=None, args=("TYPE I",)),
            Call(
                method_name="transfercmd",
                result=io.StringIO("line 1\nline 2"),
                args=("RETR dummy", None),
            ),
            Call(method_name="voidresp", result=None),
            Call(method_name="close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with host.open("dummy", "r") as fobj:
                assert fobj.closed is False
                data = fobj.readline()
                assert data == "line 1\n"
                assert fobj.closed is False
            assert fobj.closed is True

    def test_ftputil_exception(self):
        """
        If an `ftplib.FTP` method raises an exception, the `FTPFile` context
        manager should try to close the file.
        """
        host_script = [
            Call(method_name="__init__"),
            Call(method_name="pwd", result="/"),
            Call(method_name="close"),
        ]
        file_script = [
            Call(method_name="__init__"),
            Call(method_name="pwd", result="/"),
            Call(method_name="cwd", result=None, args=("/",)),
            Call(method_name="voidcmd", result=None, args=("TYPE I",)),
            # Raise exception. `voidresp` therefore won't be called, but
            # `close` will be called by the context manager.
            Call(
                method_name="transfercmd",
                result=ftplib.error_perm,
                args=("STOR inaccessible", None),
            ),
            # Call(method_name="voidresp", result=None),
            Call(method_name="close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            with pytest.raises(ftputil.error.FTPIOError):
                # This should fail.
                with host.open("/inaccessible", "w") as fobj:
                    pass
            # The file construction shouldn't have succeeded, so `fobj` should
            # be absent from the local namespace.
            assert "fobj" not in locals()

    def test_client_code_exception(self):
        """
        If client code raises an exception in the `FTPFile` context manager
        block, the file object should be closed by the context manager.
        """
        host_script = [
            Call(method_name="__init__"),
            Call(method_name="pwd", result="/"),
            Call(method_name="close"),
        ]
        file_script = [
            Call(method_name="__init__"),
            Call(method_name="pwd", result="/"),
            Call(method_name="cwd", result=None, args=("/",)),
            Call(method_name="voidcmd", result=None, args=("TYPE I",)),
            Call(
                method_name="transfercmd",
                result=io.BytesIO(b""),
                args=("RETR dummy", None),
            ),
            Call(method_name="voidresp", result=None),
            Call(method_name="close"),
        ]
        multisession_factory = scripted_session.factory(host_script, file_script)
        with test_base.ftp_host_factory(multisession_factory) as host:
            try:
                with host.open("dummy", "r") as fobj:
                    assert fobj.closed is False
                    raise ClientCodeException()
            except ClientCodeException:
                assert fobj.closed is True
            else:
                pytest.fail("`ClientCodeException` not raised")



# Copyright (C) 2006-2018, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
Provide version information about ftputil and the runtime environment.
"""

import sys


# ftputil version number; substituted by `make patch`
__version__ = "5.0.4"

_ftputil_version = __version__
_python_version = sys.version.split()[0]
_python_platform = sys.platform


version_info = "ftputil {}, Python {} ({})".format(
    _ftputil_version, _python_version, _python_platform
)



# Copyright (C) 2013-2021, Stefan Schwarzer
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
tool.py - helper code
"""

import os


__all__ = ["same_string_type_as", "as_str", "as_str_path", "raise_for_empty_path"]


def same_string_type_as(type_source, string, encoding):
    """
    Return a string of the same type as `type_source` with the content from
    `string`.

    If the `type_source` and `string` don't have the same type, use `encoding`
    to encode or decode, whatever operation is needed.
    """
    if isinstance(type_source, bytes) and isinstance(string, str):
        return string.encode(encoding)
    elif isinstance(type_source, str) and isinstance(string, bytes):
        return string.decode(encoding)
    else:
        return string


def as_str(string, encoding):
    """
    Return the argument `string` converted to a unicode string if it's a
    `bytes` object. Otherwise just return the string.

    If a conversion is necessary, use `encoding`.

    If `string` is neither `str` nor `bytes`, raise a `TypeError`.
    """
    if isinstance(string, bytes):
        return string.decode(encoding)
    elif isinstance(string, str):
        return string
    else:
        raise TypeError("`as_str` argument must be `bytes` or `str`")


def as_str_path(path, encoding):
    """
    Return the argument `path` converted to a unicode string if it's a `bytes`
    object. Otherwise just return the string.

    If a conversion is necessary, use `encoding`.

    Instead of passing a `bytes` or `str` object for `path`, you can pass a
    `PathLike` object that can be converted to a `bytes` or `str` object.

    If the `path` can't be converted to a `bytes` or `str`, a `TypeError` is
    raised.
    """
    path = os.fspath(path)
    return as_str(path, encoding)


def raise_for_empty_path(path, path_argument_name="path"):
    """
    Raise an exception of class `exception_class` if `path` is an empty string
    (text or bytes).
    """
    # Avoid cyclic import.
    import ftputil.error

    # Don't handle `pathlib.Path("")`. This immediately results in `Path(".")`,
    # so we can't detect it anyway. Regarding bytes, `Path(b"")` results in a
    # `TypeError`.
    if path in ["", b""]:
        if path_argument_name is None:
            message = "path argument is empty"
        else:
            message = f"path argument `{path_argument_name}` is empty"
        raise ftputil.error.FTPIOError(message)



# Copyright (C) 2013-2020, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
file_transfer.py - upload, download and generic file copy
"""

import os

import ftputil.stat


# TODO: Think a bit more about the API before making it public.
# Only `chunks` should be used by clients of the ftputil library. Any other
# functionality is supposed to be used via `FTPHost` objects.
# __all__ = ["chunks"]
__all__ = []

# Maximum size of chunk in `FTPHost.copyfileobj` in bytes.
MAX_COPY_CHUNK_SIZE = 64 * 1024


class LocalFile:
    """
    Represent a file on the local side which is to be transferred or is already
    transferred.
    """

    def __init__(self, name, mode):
        self.name = os.path.abspath(name)
        self.mode = mode

    def exists(self):
        """
        Return `True` if the path representing this file exists. Otherwise
        return `False`.
        """
        return os.path.exists(self.name)

    def mtime(self):
        """
        Return the timestamp for the last modification in seconds.
        """
        return os.path.getmtime(self.name)

    def mtime_precision(self):
        """
        Return the precision of the last modification time in seconds.
        """
        # Derived classes might want to use `self`.
        # pylint: disable=no-self-use
        #
        # Assume modification timestamps for local file systems are at least
        # precise up to a second.
        return 1.0

    def fobj(self):
        """
        Return a file object for the name/path in the constructor.
        """
        return open(self.name, self.mode)


class RemoteFile:
    """
    Represent a file on the remote side which is to be transferred or is
    already transferred.
    """

    def __init__(self, ftp_host, name, mode):
        self._host = ftp_host
        self._path = ftp_host.path
        self.name = self._path.abspath(name)
        self.mode = mode

    def exists(self):
        """
        Return `True` if the path representing this file exists. Otherwise
        return `False`.
        """
        return self._path.exists(self.name)

    def mtime(self):
        """
        Return the timestamp for the last modification in seconds.
        """
        # Convert to client time zone (see definition of time shift in
        # docstring of `FTPHost.set_time_shift`).
        return self._path.getmtime(self.name)

    def mtime_precision(self):
        """
        Return the precision of the last modification time in seconds.
        """
        # I think using `stat` instead of `lstat` makes more sense here.
        return self._host.stat(self.name)._st_mtime_precision

    def fobj(self):
        """
        Return a file object for the name/path in the constructor.
        """
        return self._host.open(self.name, self.mode)


def source_is_newer_than_target(source_file, target_file):
    """
    Return `True` if the source is newer than the target, else `False`.

    Both arguments are `LocalFile` or `RemoteFile` objects.

    It's assumed that the actual modification time is

      reported_mtime <= actual_mtime <= reported_mtime + mtime_precision

    i. e. that the reported mtime is the actual mtime or rounded down
    (truncated).

    For the purpose of this test the source is newer than the target if any of
    the possible actual source modification times is greater than the reported
    target modification time. In other words: If in doubt, the file should be
    transferred.

    This is the only situation where the source is _not_ considered newer than
    the target:

    |/////////////////////|              possible source mtime
                            |////////|   possible target mtime

    That is, the latest possible actual source modification time is before the
    first possible actual target modification time.
    """
    if source_file.mtime_precision() is ftputil.stat.UNKNOWN_PRECISION:
        return True
    else:
        return (
            source_file.mtime() + source_file.mtime_precision() >= target_file.mtime()
        )


def chunks(fobj, max_chunk_size=MAX_COPY_CHUNK_SIZE):
    """
    Return an iterator which yields the contents of the file object.

    For each iteration, at most `max_chunk_size` bytes are read from `fobj` and
    yielded as a byte string. If the file object is exhausted, then don't yield
    any more data but stop the iteration, so the client does _not_ get an empty
    byte string.

    Any exceptions resulting from reading the file object are passed through to
    the client.
    """
    while True:
        chunk = fobj.read(max_chunk_size)
        if not chunk:
            break
        yield chunk


def copyfileobj(
    source_fobj, target_fobj, max_chunk_size=MAX_COPY_CHUNK_SIZE, callback=None
):
    """
    Copy data from file-like object source to file-like object target.
    """
    # Inspired by `shutil.copyfileobj` (I don't use the `shutil` code directly
    # because it might change)
    for chunk in chunks(source_fobj, max_chunk_size):
        target_fobj.write(chunk)
        if callback is not None:
            callback(chunk)


def copy_file(source_file, target_file, conditional, callback):
    """
    Copy a file from `source_file` to `target_file`.

    These are `LocalFile` or `RemoteFile` objects. Which of them is a local or
    a remote file, respectively, is determined by the arguments. If
    `conditional` is true, the file is only copied if the target doesn't exist
    or is older than the source. If `conditional` is false, the file is copied
    unconditionally. Return `True` if the file was copied, else `False`.
    """
    if conditional:
        # Evaluate condition: The target file either doesn't exist or is older
        # than the source file. If in doubt (due to imprecise timestamps),
        # perform the transfer.
        transfer_condition = not target_file.exists() or source_is_newer_than_target(
            source_file, target_file
        )
        if not transfer_condition:
            # We didn't transfer.
            return False
    source_fobj = source_file.fobj()
    try:
        target_fobj = target_file.fobj()
        try:
            copyfileobj(source_fobj, target_fobj, callback=callback)
        finally:
            target_fobj.close()
    finally:
        source_fobj.close()
    # Transfer accomplished
    return True



# Copyright (C) 2003-2021, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import datetime
import ftplib
import functools
import time

import pytest

import ftputil
import ftputil.error
import ftputil.path_encoding
import ftputil.tool

from test import test_base
from test import scripted_session


Call = scripted_session.Call


def as_bytes(string, encoding=ftputil.path_encoding.DEFAULT_ENCODING):
    return string.encode(encoding)


class TestPath:
    """Test operations in `FTPHost.path`."""

    # TODO: Add unit tests for changes for ticket #113 (commits [b4c9b089b6b8]
    # and [4027740cdd2d]).
    def test_regular_isdir_isfile_islink(self):
        """
        Test regular `FTPHost._Path.isdir/isfile/islink`.
        """
        # Test a path which isn't there.
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # `isdir` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            # `isfile` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            # `islink` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert not host.path.isdir("notthere")
            assert not host.path.isfile("notthere")
            assert not host.path.islink("notthere")
        # This checks additional code (see ticket #66).
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # `isdir` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            # `isfile` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            # `islink` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert not host.path.isdir("/notthere/notthere")
            assert not host.path.isfile("/notthere/notthere")
            assert not host.path.islink("/notthere/notthere")
        # Test a directory.
        test_dir = "/some_dir"
        dir_line = test_base.dir_line(
            mode_string="dr-xr-xr-x",
            datetime_=datetime.datetime.now(),
            name=test_dir.lstrip("/"),
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # `isdir` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            # `isfile` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            # `islink` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert host.path.isdir(test_dir)
            assert not host.path.isfile(test_dir)
            assert not host.path.islink(test_dir)
        # Test a file.
        test_file = "/some_file"
        dir_line = test_base.dir_line(
            datetime_=datetime.datetime.now(), name=test_file.lstrip("/")
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # `isdir` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            # `isfile` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            # `islink` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert not host.path.isdir(test_file)
            assert host.path.isfile(test_file)
            assert not host.path.islink(test_file)
        # Test a link. Since the link target doesn't exist, neither
        # `isdir` nor `isfile` return `True`.
        test_link = "/some_link"
        dir_line = test_base.dir_line(
            mode_string="lrwxrwxrwx",
            datetime_=datetime.datetime.now(),
            name=test_link.lstrip("/"),
            link_target="nonexistent",
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # `isdir` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            #  Look for `/some_link`
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            #  Look for `/nonexistent`
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            # `isfile` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            #  Look for `/some_link`
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            #  Look for `/nonexistent`
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            # `islink` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            #  Look for `/some_link`. `islink` doesn't try to dereference
            #  the link.
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert not host.path.isdir(test_link)
            assert not host.path.isfile(test_link)
            assert host.path.islink(test_link)

    def test_workaround_for_spaces(self):
        """
        Test whether the workaround for space-containing paths is used.
        """
        # Test a file name containing spaces.
        test_file = "/home/dir with spaces/file with spaces"
        dir_line1 = test_base.dir_line(
            mode_string="dr-xr-xr-x", datetime_=datetime.datetime.now(), name="home"
        )
        dir_line2 = test_base.dir_line(
            mode_string="dr-xr-xr-x",
            datetime_=datetime.datetime.now(),
            name="dir with spaces",
        )
        dir_line3 = test_base.dir_line(
            mode_string="-r--r--r--",
            datetime_=datetime.datetime.now(),
            name="file with spaces",
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # `isdir` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line1),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/home",)),
            Call("dir", args=("",), result=dir_line2),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/home/dir with spaces",)),
            Call("dir", args=("",), result=dir_line3),
            Call("cwd", args=("/",)),
            # `isfile` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line1),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/home",)),
            Call("dir", args=("",), result=dir_line2),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/home/dir with spaces",)),
            Call("dir", args=("",), result=dir_line3),
            Call("cwd", args=("/",)),
            # `islink` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line1),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/home",)),
            Call("dir", args=("",), result=dir_line2),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/home/dir with spaces",)),
            Call("dir", args=("",), result=dir_line3),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert not host.path.isdir(test_file)
            assert host.path.isfile(test_file)
            assert not host.path.islink(test_file)

    def test_inaccessible_home_directory_and_whitespace_workaround(self):
        """
        Test combination of inaccessible home directory + whitespace in path.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", result=ftplib.error_perm),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            with pytest.raises(ftputil.error.InaccessibleLoginDirError):
                host._dir("/home dir")

    def test_isdir_isfile_islink_with_dir_failure(self):
        """
        Test failing `FTPHost._Path.isdir/isfile/islink` because of failing
        `_dir` call.
        """
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=ftplib.error_perm),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        FTPOSError = ftputil.error.FTPOSError
        # Test if exceptions are propagated.
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            with pytest.raises(FTPOSError):
                host.path.isdir("index.html")
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            with pytest.raises(FTPOSError):
                host.path.isfile("index.html")
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            with pytest.raises(FTPOSError):
                host.path.islink("index.html")

    def test_isdir_isfile_with_infinite_link_chain(self):
        """
        Test if `isdir` and `isfile` return `False` if they encounter an
        infinite link chain.
        """
        # `/home/bad_link` links to `/home/subdir/bad_link`, which links back
        # to `/home/bad_link` etc.
        dir_line1 = test_base.dir_line(
            mode_string="dr-xr-xr-x", datetime_=datetime.datetime.now(), name="home"
        )
        dir_line2 = test_base.dir_line(
            mode_string="lrwxrwxrwx",
            datetime_=datetime.datetime.now(),
            name="bad_link",
            link_target="subdir/bad_link",
        )
        dir_line3 = test_base.dir_line(
            mode_string="dr-xr-xr-x", datetime_=datetime.datetime.now(), name="subdir"
        )
        dir_line4 = test_base.dir_line(
            mode_string="lrwxrwxrwx",
            datetime_=datetime.datetime.now(),
            name="bad_link",
            link_target="/home/bad_link",
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line1),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/home",)),
            Call("dir", args=("",), result=dir_line2 + "\n" + dir_line3),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/home/subdir",)),
            Call("dir", args=("",), result=dir_line4),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            assert host.path.isdir("/home/bad_link") is False
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            assert host.path.isfile("/home/bad_link") is False

    def test_exists(self):
        """
        Test `FTPHost.path.exists`.
        """
        # Regular use of `exists`
        dir_line1 = test_base.dir_line(
            datetime_=datetime.datetime.now(), name="some_file"
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # `exists("some_file")`
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line1),
            Call("cwd", args=("/",)),
            # `exists("notthere")`
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line1),
            Call("cwd", args=("/",)),
            # `exists` with failing `dir` call
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=ftplib.error_perm),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert host.path.exists("some_file")
            assert not host.path.exists("notthere")
            # Test if exceptions are propagated.
            with pytest.raises(ftputil.error.FTPOSError):
                host.path.exists("some_file")


class TestAcceptEitherBytesOrStr:

    # Use path arguments directly
    path_converter = staticmethod(lambda path: path)

    def _test_method_string_types(self, method, path):
        expected_type = type(path)
        path_converter = self.path_converter
        assert isinstance(method(path_converter(path)), expected_type)

    def test_methods_that_take_and_return_one_string(self):
        """
        Test whether the same string type as for the argument is returned.
        """
        method_names = [
            "abspath",
            "basename",
            "dirname",
            "join",
            "normcase",
            "normpath",
        ]
        script = [Call("__init__"), Call("pwd", result="/"), Call("close")]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            for method_name in method_names:
                method = getattr(host.path, method_name)
                self._test_method_string_types(method, "/")
                self._test_method_string_types(method, ".")
                self._test_method_string_types(method, b"/")
                self._test_method_string_types(method, b".")

    def test_methods_that_take_a_string_and_return_a_bool(self):
        """
        Test whether the methods accept byte and unicode strings.
        """
        path_converter = self.path_converter
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # `exists` test 1
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(name="ä", datetime_=datetime.datetime.now()),
            ),
            Call("cwd", args=("/",)),
            # `exists` test 2
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(name="ä", datetime_=datetime.datetime.now()),
            ),
            Call("cwd", args=("/",)),
            # `isdir` test 1
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(
                    mode_string="dr-xr-xr-x",
                    name="ä",
                    datetime_=datetime.datetime.now(),
                ),
            ),
            Call("cwd", args=("/",)),
            # `isdir` test 2
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(
                    mode_string="dr-xr-xr-x",
                    name="ä",
                    datetime_=datetime.datetime.now(),
                ),
            ),
            Call("cwd", args=("/",)),
            # `isfile` test 1
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(name="ö", datetime_=datetime.datetime.now()),
            ),
            Call("cwd", args=("/",)),
            # `isfile` test 2
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(name="ö", datetime_=datetime.datetime.now()),
            ),
            Call("cwd", args=("/",)),
            # `islink` test 1
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(
                    mode_string="lrwxrwxrwx",
                    name="ü",
                    datetime_=datetime.datetime.now(),
                    link_target="unimportant",
                ),
            ),
            Call("cwd", args=("/",)),
            # `islink` test 2
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(
                    mode_string="lrwxrwxrwx",
                    name="ü",
                    datetime_=datetime.datetime.now(),
                    link_target="unimportant",
                ),
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        custom_as_bytes = functools.partial(
            as_bytes, encoding=ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING
        )
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            # `isabs`
            assert not host.path.isabs("ä")
            assert not host.path.isabs(path_converter(custom_as_bytes("ä")))
            # `exists`
            assert host.path.exists(path_converter("ä"))
            assert host.path.exists(path_converter(custom_as_bytes("ä")))
            # `isdir`, `isfile`, `islink`
            assert host.path.isdir(path_converter("ä"))
            assert host.path.isdir(path_converter(custom_as_bytes("ä")))
            assert host.path.isfile(path_converter("ö"))
            assert host.path.isfile(path_converter(custom_as_bytes("ö")))
            assert host.path.islink(path_converter("ü"))
            assert host.path.islink(path_converter(custom_as_bytes("ü")))

    def test_getmtime(self):
        """
        Test whether `FTPHost.path.getmtime` accepts byte and unicode paths.
        """
        path_converter = self.path_converter
        now = datetime.datetime.utcnow()
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # `getmtime` call 1
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=test_base.dir_line(name="ä", datetime_=now)),
            Call("cwd", args=("/",)),
            # `getmtime` call 2
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=test_base.dir_line(name="ä", datetime_=now)),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        expected_mtime = now.timestamp()
        # We don't care about the _exact_ time, so don't bother with timezone
        # differences. Instead, do a simple sanity check.
        day = 24 * 60 * 60  # seconds
        mtime_makes_sense = (
            lambda mtime: expected_mtime - day <= mtime <= expected_mtime + day
        )
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert mtime_makes_sense(host.path.getmtime(path_converter(("ä"))))
            assert mtime_makes_sense(
                host.path.getmtime(
                    path_converter(
                        as_bytes("ä", ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING)
                    )
                )
            )

    def test_getsize(self):
        """
        Test whether `FTPHost.path.getsize` accepts byte and unicode paths.
        """
        path_converter = self.path_converter
        now = datetime.datetime.now()
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # `getsize` call 1
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(name="ä", size=512, datetime_=now),
            ),
            Call("cwd", args=("/",)),
            # `getsize` call 2
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(name="ä", size=512, datetime_=now),
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            assert host.path.getsize(path_converter("ä")) == 512
            assert (
                host.path.getsize(
                    path_converter(
                        as_bytes("ä", ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING)
                    )
                )
                == 512
            )

    def test_walk(self):
        """
        Test whether `FTPHost.path.walk` accepts bytes and unicode paths.
        """
        path_converter = self.path_converter
        now = datetime.datetime.now()
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            # `walk` call 1
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(
                    mode_string="dr-xr-xr-x", name="ä", size=512, datetime_=now
                ),
            ),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/ä",)),
            #  Assume directory `ä` is empty.
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            # `walk` call 2
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result=test_base.dir_line(
                    mode_string="dr-xr-xr-x", name="ä", size=512, datetime_=now
                ),
            ),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/ä",)),
            #  Assume directory `ä` is empty.
            Call("dir", args=("",), result=""),
            Call("cwd", args=("/",)),
            Call("close"),
        ]

        def noop(arg, top, names):
            del names[:]

        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.stat_cache.disable()
            host.path.walk(path_converter("ä"), func=noop, arg=None)
            host.path.walk(
                path_converter(
                    as_bytes("ä", ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING)
                ),
                func=noop,
                arg=None,
            )


class Path:
    def __init__(self, path):
        self.path = path

    def __fspath__(self):
        return self.path


class TestAcceptEitherBytesOrStrFromPath(TestAcceptEitherBytesOrStr):

    # Take path arguments from `Path(...)` objects
    path_converter = Path



# Copyright (C) 2002-2020, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import ftplib
import socket

import pytest

import ftputil.error


class TestFTPErrorArguments:
    """
    The `*Error` constructors should accept either a byte string or a unicode
    string.
    """

    def test_bytestring_argument(self):
        # An umlaut as latin-1 character
        io_error = ftputil.error.FTPIOError(b"\xe4")
        os_error = ftputil.error.FTPOSError(b"\xe4")

    def test_unicode_argument(self):
        # An umlaut as unicode character
        io_error = ftputil.error.FTPIOError("\xe4")
        os_error = ftputil.error.FTPOSError("\xe4")


class TestErrorConversion:
    def callee(self):
        raise ftplib.error_perm()

    def test_ftplib_error_to_ftp_os_error(self):
        """
        Ensure the `ftplib` exception isn't used as `FTPOSError` argument.
        """
        with pytest.raises(ftputil.error.FTPOSError) as exc_info:
            with ftputil.error.ftplib_error_to_ftp_os_error:
                self.callee()
        exc = exc_info.value
        assert not (exc.args and isinstance(exc.args[0], ftplib.error_perm))
        del exc_info

    def test_ftplib_error_to_ftp_io_error(self):
        """
        Ensure the `ftplib` exception isn't used as `FTPIOError` argument.
        """
        with pytest.raises(ftputil.error.FTPIOError) as exc_info:
            with ftputil.error.ftplib_error_to_ftp_io_error:
                self.callee()
        exc = exc_info.value
        assert not (exc.args and isinstance(exc.args[0], ftplib.error_perm))
        del exc_info

    def test_error_message_reuse(self):
        """
        Test if the error message string is retained if the caught exception
        has more than one element in `args`.
        """
        # See ticket #76.
        with pytest.raises(ftputil.error.FTPOSError) as exc_info:
            # Format "host:port" doesn't work. The use here is intentional.
            host = ftputil.FTPHost("localhost:21", "", "")
        exc = exc_info.value
        assert isinstance(exc.__cause__, socket.gaierror)
        assert exc.__cause__.errno == socket.EAI_NONAME
        del exc_info



# Copyright (C) 2006-2018, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import time

import pytest

import ftputil.error
import ftputil.stat_cache

from test import scripted_session
from test import test_base


Call = scripted_session.Call


class TestStatCache:
    def setup_method(self, method):
        self.cache = ftputil.stat_cache.StatCache()

    def test_get_set(self):
        with pytest.raises(ftputil.error.CacheMissError):
            self.cache.__getitem__("/path")
        self.cache["/path"] = "test"
        assert self.cache["/path"] == "test"

    def test_invalidate(self):
        # Don't raise a `CacheMissError` for missing paths
        self.cache.invalidate("/path")
        self.cache["/path"] = "test"
        self.cache.invalidate("/path")
        assert len(self.cache) == 0

    def test_clear(self):
        self.cache["/path1"] = "test1"
        self.cache["/path2"] = "test2"
        self.cache.clear()
        assert len(self.cache) == 0

    def test_contains(self):
        self.cache["/path1"] = "test1"
        assert "/path1" in self.cache
        assert "/path2" not in self.cache

    def test_len(self):
        assert len(self.cache) == 0
        self.cache["/path1"] = "test1"
        self.cache["/path2"] = "test2"
        assert len(self.cache) == 2

    def test_resize(self):
        self.cache.resize(100)
        # Don't grow the cache beyond it's set size.
        for i in range(150):
            self.cache["/{0:d}".format(i)] = i
        assert len(self.cache) == 100

    def test_max_age1(self):
        """
        Set expiration after setting a cache item.
        """
        self.cache["/path1"] = "test1"
        # Expire after one second
        self.cache.max_age = 1
        time.sleep(0.5)
        # Should still be present
        assert self.cache["/path1"] == "test1"
        time.sleep(0.6)
        # Should have expired (_setting_ the cache counts)
        with pytest.raises(ftputil.error.CacheMissError):
            self.cache.__getitem__("/path1")

    def test_max_age2(self):
        """
        Set expiration before setting a cache item.
        """
        # Expire after one second
        self.cache.max_age = 1
        self.cache["/path1"] = "test1"
        time.sleep(0.5)
        # Should still be present
        assert self.cache["/path1"] == "test1"
        time.sleep(0.6)
        # Should have expired (_setting_ the cache counts)
        with pytest.raises(ftputil.error.CacheMissError):
            self.cache.__getitem__("/path1")

    def test_disabled(self):
        self.cache["/path1"] = "test1"
        self.cache.disable()
        self.cache["/path2"] = "test2"
        with pytest.raises(ftputil.error.CacheMissError):
            self.cache.__getitem__("/path1")
        with pytest.raises(ftputil.error.CacheMissError):
            self.cache.__getitem__("/path2")
        assert len(self.cache) == 1
        # Don't raise a `CacheMissError` for missing paths.
        self.cache.invalidate("/path2")

    def test_cache_size_zero(self):
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call(
                "dir",
                args=("",),
                result="drwxr-sr-x   2 45854   200    512 Jan  3 17:17 download\n"
                "drwxr-sr-x   2 45854   200    512 Jul 30 17:14 dir with spaces\n"
                "lrwxrwxrwx   2 45854   200      6 May 29  2000 link -> ../link_target\n"
                "-rw-r--r--   1 45854   200   4604 Jan 19 23:11 index.html",
            ),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            with pytest.raises(ValueError):
                host.stat_cache.resize(0)
            # If bug #38 was present, this would raise an `IndexError`.
            items = host.listdir(host.curdir)
            assert items == ["download", "dir with spaces", "link", "index.html"]



# Copyright (C) 2003-2022, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
ftputil.file - support for file-like objects on FTP servers
"""

import ftputil.error


# This module shouldn't be used by clients of the ftputil library.
__all__ = []


try:
    import ssl
except ImportError:
    SSLSocket = None
else:
    SSLSocket = ssl.SSLSocket


class FTPFile:
    """
    Represents a file-like object associated with an FTP host. File and socket
    are closed appropriately if the `close` method is called.
    """

    # Set timeout in seconds when closing file connections (see ticket #51).
    _close_timeout = 5

    def __init__(self, host):
        """Construct the file(-like) object."""
        self._host = host
        # pylint: disable=protected-access
        self._session = host._session
        # The file is still closed.
        self.closed = True
        self._conn = None
        self._fobj = None

    def _open(
        self,
        path,
        mode,
        buffering=None,
        encoding=None,
        errors=None,
        newline=None,
        *,
        rest=None,
    ):
        """
        Open the remote file with given path name and mode.

        Contrary to the `open` builtin, this method returns `None`, instead
        this file object is modified in-place.
        """
        # We use the same arguments as in `open`.
        # pylint: disable=unused-argument
        # pylint: disable=too-many-arguments
        #
        # Check mode.
        if mode is None:
            # This is Python's behavior for local files.
            raise TypeError("open() argument 2 must be str, not None")
        if "a" in mode:
            raise ftputil.error.FTPIOError("append mode not supported")
        if mode not in ("r", "rb", "rt", "w", "wb", "wt"):
            raise ftputil.error.FTPIOError("invalid mode '{}'".format(mode))
        if "b" in mode and "t" in mode:
            # Raise a `ValueError` like Python would.
            raise ValueError("can't have text and binary mode at once")
        # Convenience variables
        is_binary_mode = "b" in mode
        is_read_mode = "r" in mode
        # `rest` is only allowed for binary mode.
        if (not is_binary_mode) and (rest is not None):
            raise ftputil.error.CommandNotImplementedError(
                "`rest` argument can't be used for text files"
            )
        # Always use binary mode and leave any conversions to Python,
        # controlled by the arguments to `makefile` below.
        transfer_type = "I"
        command = "TYPE {}".format(transfer_type)
        with ftputil.error.ftplib_error_to_ftp_io_error:
            self._session.voidcmd(command)
        # Make transfer command.
        command_type = "RETR" if is_read_mode else "STOR"
        command = "{} {}".format(command_type, path)
        # Get connection and file object.
        with ftputil.error.ftplib_error_to_ftp_io_error:
            self._conn = self._session.transfercmd(command, rest)
        self._fobj = self._conn.makefile(
            mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline
        )
        # This comes last so that `close` won't try to close `FTPFile` objects
        # without `_conn` and `_fobj` attributes in case of an error.
        self.closed = False

    def __iter__(self):
        """
        Return a file iterator.
        """
        return self

    def __next__(self):
        """
        Return the next line or raise `StopIteration`, if there are no more.
        """
        # Apply implicit line ending conversion for text files.
        line = self.readline()
        if line:
            return line
        else:
            raise StopIteration

    #
    # Context manager methods
    #
    def __enter__(self):
        # Return `self`, so it can be accessed as the variable component of the
        # `with` statement.
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # We don't need the `exc_*` arguments here
        # pylint: disable=unused-argument
        self.close()
        # Be explicit
        return False

    #
    # Other attributes
    #
    def __getattr__(self, attr_name):
        """
        Handle requests for attributes unknown to `FTPFile` objects: delegate
        the requests to the contained file object.
        """
        if attr_name in (
            "encoding flush isatty fileno read readline readlines seek tell "
            "truncate name softspace write writelines".split()
        ):
            return getattr(self._fobj, attr_name)
        raise AttributeError("'FTPFile' object has no attribute '{}'".format(attr_name))

    # TODO: Implement `__dir__`? (See
    # http://docs.python.org/whatsnew/2.6.html#other-language-changes )

    def close(self):
        """
        Close the `FTPFile`.
        """
        if self.closed:
            return
        # Timeout value to restore, see below.
        # Statement works only before the try/finally statement, otherwise
        # Python raises an `UnboundLocalError`.
        old_timeout = self._session.sock.gettimeout()
        try:
            self._fobj.close()
            self._fobj = None
            with ftputil.error.ftplib_error_to_ftp_io_error:
                if (SSLSocket is not None) and isinstance(self._conn, SSLSocket):
                    self._conn.unwrap()
                self._conn.close()
            # Set a timeout to prevent waiting until server timeout if we have
            # a server blocking here like in ticket #51.
            self._session.sock.settimeout(self._close_timeout)
            try:
                with ftputil.error.ftplib_error_to_ftp_io_error:
                    self._session.voidresp()
            except ftputil.error.FTPIOError as exc:
                # Ignore some errors, see tickets #51 and #17 at
                # http://ftputil.sschwarzer.net/trac/ticket/51 and
                # http://ftputil.sschwarzer.net/trac/ticket/17, respectively.
                exc = str(exc)
                error_code = exc[:3]
                if exc.splitlines()[0] != "timed out" and error_code not in (
                    "150",
                    "426",
                    "450",
                    "451",
                ):
                    raise
        finally:
            # Restore timeout for socket of `FTPFile`'s `ftplib.FTP` object in
            # case the connection is reused later.
            self._session.sock.settimeout(old_timeout)
            # If something went wrong before, the file is probably defunct and
            # subsequent calls to `close` won't help either, so we consider the
            # file closed for practical purposes.
            self.closed = True

    def __getstate__(self):
        raise TypeError("cannot serialize FTPFile object")



# Copyright (C) 2006-2018, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
ftp_stat_cache.py - cache for (l)stat data
"""

import time

import ftputil.error
import ftputil.lrucache


# This module shouldn't be used by clients of the ftputil library.
__all__ = []


class StatCache:
    """
    Implement an LRU (least-recently-used) cache.

    `StatCache` objects have an attribute `max_age`. After this duration after
    _setting_ it a cache entry will expire. For example, if you code

    my_cache = StatCache()
    my_cache.max_age = 10
    my_cache["/home"] = ...

    the value my_cache["/home"] can be retrieved for 10 seconds. After that,
    the entry will be treated as if it had never been in the cache and should
    be fetched again from the remote host.

    Note that the `__len__` method does no age tests and thus may include some
    or many already expired entries.
    """

    # Default number of cache entries
    _DEFAULT_CACHE_SIZE = 5000

    def __init__(self):
        # Can be reset with method `resize`
        self._cache = ftputil.lrucache.LRUCache(self._DEFAULT_CACHE_SIZE)
        # Never expire
        self.max_age = None
        self.enable()

    def enable(self):
        """
        Enable storage of stat results.
        """
        self._enabled = True

    def disable(self):
        """
        Disable the cache. Further storage attempts with `__setitem__` won't
        have any visible effect.

        Disabling the cache only effects new storage attempts. Values stored
        before calling `disable` can still be retrieved unless disturbed by a
        `resize` command or normal cache expiration.
        """
        # `_enabled` is set via calling `enable` in the constructor.
        # pylint: disable=attribute-defined-outside-init
        self._enabled = False

    def resize(self, new_size):
        """
        Set number of cache entries to the integer `new_size`. If the new size
        is smaller than the current cache size, relatively long-unused elements
        will be removed.
        """
        self._cache.size = new_size

    def _age(self, path):
        """
        Return the age of a cache entry for `path` in seconds. If the path
        isn't in the cache, raise a `CacheMissError`.
        """
        try:
            return time.time() - self._cache.mtime(path)
        except ftputil.lrucache.CacheKeyError:
            raise ftputil.error.CacheMissError(
                "no entry for path {} in cache".format(path)
            )

    def clear(self):
        """
        Clear (invalidate) all cache entries.
        """
        self._cache.clear()

    def invalidate(self, path):
        """
        Invalidate the cache entry for the absolute `path` if present. After
        that, the stat result data for `path` can no longer be retrieved, as if
        it had never been stored.

        If no stat result for `path` is in the cache, do _not_ raise an
        exception.
        """
        # XXX: To be 100 % sure, this should be `host.sep`, but I don't want to
        # introduce a reference to the `FTPHost` object for only that purpose.
        assert path.startswith("/"), "{} must be an absolute path".format(path)
        try:
            del self._cache[path]
        except ftputil.lrucache.CacheKeyError:
            # Ignore errors
            pass

    def __getitem__(self, path):
        """
        Return the stat entry for the `path`. If there's no stored stat entry
        or the cache is disabled, raise `CacheMissError`.
        """
        if not self._enabled:
            raise ftputil.error.CacheMissError("cache is disabled")
        # Possibly raise a `CacheMissError` in `_age`
        if (self.max_age is not None) and (self._age(path) > self.max_age):
            self.invalidate(path)
            raise ftputil.error.CacheMissError(
                "entry for path {} has expired".format(path)
            )
        else:
            # XXX: I don't know if this may raise a `CacheMissError` in case of
            # race conditions. I prefer robust code.
            try:
                return self._cache[path]
            except ftputil.lrucache.CacheKeyError:
                raise ftputil.error.CacheMissError(
                    "entry for path {} not found".format(path)
                )

    def __setitem__(self, path, stat_result):
        """
        Put the stat data for the absolute `path` into the cache, unless it's
        disabled.
        """
        assert path.startswith("/")
        if not self._enabled:
            return
        self._cache[path] = stat_result

    def __contains__(self, path):
        """
        Support for the `in` operator. Return a true value, if data for `path`
        is in the cache, else return a false value.
        """
        try:
            # Implicitly do an age test which may raise `CacheMissError`.
            self[path]
        except ftputil.error.CacheMissError:
            return False
        else:
            return True

    #
    # The following methods are only intended for debugging!
    #
    def __len__(self):
        """
        Return the number of entries in the cache. Note that this may include
        some (or many) expired entries.
        """
        return len(self._cache)

    def __str__(self):
        """
        Return a string representation of the cache contents.
        """
        lines = []
        for key in sorted(self._cache):
            lines.append("{}: {}".format(key, self[key]))
        return "\n".join(lines)



# Copyright (C) 2014-2021, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

"""
Unit tests for session factory helpers.
"""

import ftplib
import sys

import ftputil.path_encoding
import ftputil.session
import ftputil.tool


# Inherit from `ftplib.FTP` to get past the subclass check in
# `ftputil.session.session_factory`.
class MockSession(ftplib.FTP):
    """
    Mock session base class to determine if all expected calls have happened.
    """

    encoding = ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING

    def __init__(self, encoding=None):
        self.calls = []
        if encoding is not None:
            self.encoding = encoding

    def add_call(self, *args):
        self.calls.append(args)

    def connect(self, host, port):
        self.add_call("connect", host, port)

    def set_debuglevel(self, value):
        self.add_call("set_debuglevel", value)

    def login(self, user, password):
        self.add_call("login", user, password)

    def set_pasv(self, flag):
        self.add_call("set_pasv", flag)


class EncryptedMockSession(MockSession):
    def auth_tls(self):
        self.add_call("auth_tls")

    def prot_p(self):
        self.add_call("prot_p")


class TestSessionFactory:
    """
    Test if session factories created by `ftputil.session.session_factory`
    trigger the expected calls.
    """

    def test_defaults(self):
        """
        Test defaults (apart from base class).
        """
        factory = ftputil.session.session_factory(base_class=MockSession)
        session = factory("host", "user", "password")
        assert session.calls == [("connect", "host", 21), ("login", "user", "password")]

    def test_different_port(self):
        """
        Test setting the command channel port with `port`.
        """
        factory = ftputil.session.session_factory(base_class=MockSession, port=2121)
        session = factory("host", "user", "password")
        assert session.calls == [
            ("connect", "host", 2121),
            ("login", "user", "password"),
        ]

    def test_use_passive_mode(self):
        """
        Test explicitly setting passive/active mode with `use_passive_mode`.
        """
        # Passive mode
        factory = ftputil.session.session_factory(
            base_class=MockSession, use_passive_mode=True
        )
        session = factory("host", "user", "password")
        assert session.calls == [
            ("connect", "host", 21),
            ("login", "user", "password"),
            ("set_pasv", True),
        ]
        # Active mode
        factory = ftputil.session.session_factory(
            base_class=MockSession, use_passive_mode=False
        )
        session = factory("host", "user", "password")
        assert session.calls == [
            ("connect", "host", 21),
            ("login", "user", "password"),
            ("set_pasv", False),
        ]

    def test_encrypt_data_channel(self):
        """
        Test request to call `prot_p` with `encrypt_data_channel`.
        """
        # With encrypted data channel (default for encrypted session).
        factory = ftputil.session.session_factory(base_class=EncryptedMockSession)
        session = factory("host", "user", "password")
        assert session.calls == [
            ("connect", "host", 21),
            ("login", "user", "password"),
            ("prot_p",),
        ]
        #
        factory = ftputil.session.session_factory(
            base_class=EncryptedMockSession, encrypt_data_channel=True
        )
        session = factory("host", "user", "password")
        assert session.calls == [
            ("connect", "host", 21),
            ("login", "user", "password"),
            ("prot_p",),
        ]
        # Without encrypted data channel.
        factory = ftputil.session.session_factory(
            base_class=EncryptedMockSession, encrypt_data_channel=False
        )
        session = factory("host", "user", "password")
        assert session.calls == [("connect", "host", 21), ("login", "user", "password")]

    def test_encoding(self):
        """
        Test setting the default encoding and a custom encoding.
        """
        # Default encoding
        factory = ftputil.session.session_factory(
            base_class=MockSession,
        )
        session = factory("host", "user", "password")
        assert session.calls == [
            ("connect", "host", 21),
            ("login", "user", "password"),
        ]
        assert session.encoding == ftputil.path_encoding.FTPLIB_DEFAULT_ENCODING
        # Custom encoding
        factory = ftputil.session.session_factory(
            base_class=MockSession,
            encoding="UTF-8",
        )
        session = factory("host", "user", "password")
        assert session.calls == [
            ("connect", "host", 21),
            ("login", "user", "password"),
        ]
        assert session.encoding == "UTF-8"

    def test_debug_level(self):
        """
        Test setting the debug level on the session.
        """
        factory = ftputil.session.session_factory(base_class=MockSession, debug_level=1)
        session = factory("host", "user", "password")
        assert session.calls == [
            ("connect", "host", 21),
            ("set_debuglevel", 1),
            ("login", "user", "password"),
        ]



# Copyright (C) 2010-2020, Stefan Schwarzer <sschwarzer@sschwarzer.net>
# and ftputil contributors (see `doc/contributors.txt`)
# See the file LICENSE for licensing terms.

import datetime
import io
import random

import pytest

import ftputil.file_transfer
import ftputil.stat

from test import test_base
from test import scripted_session


Call = scripted_session.Call


class MockFile:
    """
    Class compatible with `LocalFile` and `RemoteFile`.
    """

    def __init__(self, mtime, mtime_precision):
        self._mtime = mtime
        self._mtime_precision = mtime_precision

    def mtime(self):
        return self._mtime

    def mtime_precision(self):
        return self._mtime_precision


class TestRemoteFile:
    def test_time_shift_subtracted_only_once(self):
        """
        Test whether the time shift value is subtracted from the initial server
        timestamp only once.

        This subtraction happens in `stat._Stat.parse_unix_time`, so it must
        _not_ be done a second time in `file_transfer.RemoteFile`.
        """
        utcnow = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc)
        # 3 hours
        time_shift = 3 * 60 * 60
        dir_line = test_base.dir_line(
            datetime_=utcnow + datetime.timedelta(seconds=time_shift), name="dummy_name"
        )
        script = [
            Call("__init__"),
            Call("pwd", result="/"),
            Call("cwd", args=("/",)),
            Call("cwd", args=("/",)),
            Call("dir", args=("",), result=dir_line),
            Call("cwd", args=("/",)),
            Call("close"),
        ]
        with test_base.ftp_host_factory(scripted_session.factory(script)) as host:
            host.set_time_shift(3 * 60 * 60)
            remote_file = ftputil.file_transfer.RemoteFile(host, "dummy_name", 0o644)
            remote_mtime = remote_file.mtime()
        # The remote mtime should be corrected by the time shift, so the
        # calculated UTC time is the same as for the client. The 60.0 (seconds)
        # is the timestamp precision.
        assert remote_mtime <= utcnow.timestamp() <= remote_mtime + 60.0


class TestTimestampComparison:
    def test_source_is_newer_than_target(self):
        """
        Test whether the source is newer than the target, i. e. if the
        file should be transferred.
        """
        # Define some time units/precisions.
        second = 1.0
        minute = 60 * second
        hour = 60 * minute
        day = 24 * hour
        unknown = ftputil.stat.UNKNOWN_PRECISION
        # Define input arguments; modification datetimes are in seconds. Fields
        # are source datetime/precision, target datetime/precision, expected
        # comparison result.
        file_data = [
            # Non-overlapping modification datetimes/precisions
            (1000.0, second, 900.0, second, True),
            (900.0, second, 1000.0, second, False),
            # Equal modification datetimes/precisions (if in doubt, transfer)
            (1000.0, second, 1000.0, second, True),
            # Just touching intervals
            (1000.0, second, 1000.0 + second, minute, True),
            (1000.0 + second, minute, 1000.0, second, True),
            # Other overlapping intervals
            (10000.0 - 0.5 * hour, hour, 10000.0, day, True),
            (10000.0 + 0.5 * hour, hour, 10000.0, day, True),
            (10000.0 + 0.2 * hour, 0.2 * hour, 10000.0, hour, True),
            (10000.0 - 0.2 * hour, 2 * hour, 10000.0, hour, True),
            # Unknown precision
            (1000.0, unknown, 1000.0, second, True),
            (1000.0, second, 1000.0, unknown, True),
            (1000.0, unknown, 1000.0, unknown, True),
        ]
        for (
            source_mtime,
            source_mtime_precision,
            target_mtime,
            target_mtime_precision,
            expected_result,
        ) in file_data:
            source_file = MockFile(source_mtime, source_mtime_precision)
            target_file = MockFile(target_mtime, target_mtime_precision)
            result = ftputil.file_transfer.source_is_newer_than_target(
                source_file, target_file
            )
            assert result == expected_result


class FailingStringIO(io.BytesIO):
    """
    Mock class to test whether exceptions are passed on.
    """

    # Kind of nonsense; we just want to see this exception raised.
    expected_exception = IndexError

    def read(self, count):
        raise self.expected_exception


class TestChunkwiseTransfer:
    def _random_string(self, count):
        """
        Return a `BytesIO` object containing `count` "random" bytes.
        """
        ints = (random.randint(0, 255) for i in range(count))
        return bytes(ints)

    def test_chunkwise_transfer_without_remainder(self):
        """
        Check if we get four chunks with 256 Bytes each.
        """
        data = self._random_string(1024)
        fobj = io.BytesIO(data)
        chunks = list(ftputil.file_transfer.chunks(fobj, 256))
        assert len(chunks) == 4
        assert chunks[0] == data[:256]
        assert chunks[1] == data[256:512]
        assert chunks[2] == data[512:768]
        assert chunks[3] == data[768:1024]

    def test_chunkwise_transfer_with_remainder(self):
        """
        Check if we get three chunks with 256 Bytes and one with 253.
        """
        data = self._random_string(1021)
        fobj = io.BytesIO(data)
        chunks = list(ftputil.file_transfer.chunks(fobj, 256))
        assert len(chunks) == 4
        assert chunks[0] == data[:256]
        assert chunks[1] == data[256:512]
        assert chunks[2] == data[512:768]
        assert chunks[3] == data[768:1021]

    def test_chunkwise_transfer_with_exception(self):
        """
        Check if we see the exception raised during reading.
        """
        data = self._random_string(1024)
        fobj = FailingStringIO(data)
        iterator = ftputil.file_transfer.chunks(fobj, 256)
        with pytest.raises(FailingStringIO.expected_exception):
            next(iterator)
